from matplotlib import pyplot as plt

from .time_series_visualisation import Visualise, VisualiseKMeans, VisualisePCA, VisualiseUMAP
from ..metrics.losses import PVarianceLoss


def UMAP_PCA_report(pred_viz: Visualise, target_viz: Visualise, **kwargs):
	fig, axes = plt.subplots(ncols=2, nrows=3, figsize=kwargs.get("figsize", (16, 8)))
	Visualise.number_axes(axes)
	viz_pca_target = Visualise(
		timeseries=target_viz.timeseries,
		shape=target_viz.shape,
	)
	viz_pca = VisualisePCA(
		timeseries=pred_viz.timeseries,
		shape=pred_viz.shape,
	).trajectory_pca(target=viz_pca_target, fig=fig, axes=axes[:, 1], show=False)
	viz_umap_target = Visualise(
		timeseries=target_viz.timeseries,
		shape=target_viz.shape,
	)
	viz_umap = VisualiseUMAP(
		timeseries=pred_viz.timeseries,
		shape=pred_viz.shape,
	).trajectory_umap(
		UMAPs=(1, 2), target=viz_umap_target,
		fig=fig, axes=axes[:, 0],
		filename=kwargs.get("filename", None),
		show=kwargs.get("show", True)
	)


def mix_report(pred_viz: Visualise, target_viz: Visualise, **kwargs):
	fig, axes = plt.subplots(ncols=2, nrows=3, figsize=kwargs.get("figsize", (16, 8)))
	Visualise.number_axes(axes)
	pVar = PVarianceLoss()(pred_viz.timeseries, target_viz.timeseries)
	title = f"Network Prediction (pVar: {pVar:.3f})"
	fig.suptitle(title, fontsize=kwargs.get("fontsize", 16))
	
	pred_viz.plot_timeseries_comparison(
		target_viz.timeseries, kwargs.get("spikes"),
		n_spikes_steps=kwargs.get("n_spikes_steps"),
		title="", desc=kwargs.get("desc", "Prediction"),
		fig=fig, axes=axes[:, 0],
		traces_to_show=kwargs.get("traces_to_show", ["best", "most_var", "typical_0"]),
		traces_to_show_names=kwargs.get("traces_to_show_names", [
			"Best Neuron Prediction", "Most variable Neuron Prediction", "Typical Neuron Prediction"
		]),
		show=False,
	)
	space = kwargs.get("space", "UMAP")
	if isinstance(pred_viz, (VisualiseUMAP, VisualisePCA)):
		pred_viz_space = pred_viz
	elif space == "UMAP":
		pred_viz_space = VisualiseUMAP(
			timeseries=pred_viz.timeseries,
			shape=pred_viz.shape,
		)
	elif space == "PCA":
		pred_viz_space = VisualisePCA(
			timeseries=pred_viz.timeseries,
			shape=pred_viz.shape,
		)
	else:
		raise ValueError(f"Unknown space: {space}. Try 'UMAP' or 'PCA'.")
	pred_viz_space.trajectory_umap(
		UMAPs=(1, 2), target=target_viz,
		fig=fig, axes=axes[:, 1],
		filename=kwargs.get("filename", None),
		show=kwargs.get("show", True)
	)







from .spike_funcs import (
	SpikeFuncType,
	SpikeFunction,
	SpikeFuncType2Func,
	HeavisideSigmoidApprox,
	HeavisidePhiApprox,
)

from .layers import (
	LayerType,
	BaseLayer,
	Linear,
	LIFLayer,
	ALIFLayer,
	IzhikevichLayer,
	LILayer,
	LayerType2Layer,
	SpyLIFLayer,
	SpyALIFLayer,
	SpyLILayer,
)


from .base import (
	BaseModel,
)

from .sequential_rnn import (
	SequentialRNN,
)



import torch


class WeirdTanh(torch.nn.Module):
	r"""
	Applies the Hyperbolic Tangent (Tanh) function element-wise.
	
	Tanh is defined as:
	
	.. math::
		\text{Tanh}(x) = \tanh(x) = \frac{a\exp(\alpha x) - b\exp(-\beta x)} {c\exp(\gamma x) + d\exp(-\delta x)}
	"""
	def __init__(
			self,
			a: float = 1.0, b: float = 1.0, c: float = 1.0, d: float = 1.0,
			alpha: float = 1.0, beta: float = 1.0, gamma: float = 1.0, delta: float = 1.0,
	):
		super().__init__()
		self.a = a
		self.b = b
		self.c = c
		self.d = d
		self.alpha = alpha
		self.beta = beta
		self.gamma = gamma
		self.delta = delta
	
	def forward(self, x):
		numerator = self.a * torch.exp(self.alpha * x) - self.b * torch.exp(-self.beta * x)
		denominator = self.c * torch.exp(self.gamma * x) + self.d * torch.exp(-self.delta * x)
		return numerator / denominator


"""
Goal: Make e-prop with auto-differentiation.
References paper:
	- https://arxiv.org/pdf/2201.07602.pdf
	- RFLO: https://elifesciences.org/articles/43299#s4
	- Bellec: https://www.biorxiv.org/content/10.1101/738385v3.full.pdf+html
	
References code:
	- https://github.com/ChFrenkel/eprop-PyTorch/blob/main/models.py
	- Bellec: https://github.com/IGITUGraz/eligibility_propagation
"""
from typing import Optional

import torch
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm

from neurotorch import to_numpy, to_tensor
from neurotorch.dimension import SizeTypes
from neurotorch.metrics import PVarianceLoss
from neurotorch.modules.layers import Linear, LILayer, WilsonCowanLayer, SpyLILayer, BellecLIFLayer
from neurotorch.utils import unpack_out_hh, legend_without_duplicate_labels_, batchwise_temporal_recursive_filter


def compute_dz_dw_local(z, w):
	z_view = z.view(-1)
	dzdw = torch.zeros_like(w)
	grad_outputs = torch.eye(z_view.shape[-1], device=w.device)
	for g_idx in range(grad_outputs.shape[0]):
		w.grad = torch.zeros_like(w)
		dzdw[g_idx] = torch.autograd.grad(z_view[g_idx], w, retain_graph=True)[0][g_idx]
	return dzdw


class DummyLayer(Linear):
	def __init__(
			self,
			input_size: Optional[SizeTypes] = None,
			output_size: Optional[SizeTypes] = None,
			name: Optional[str] = None,
			device: Optional[torch.device] = None,
			**kwargs
	):
		super().__init__(
			input_size=input_size,
			output_size=output_size,
			name=name,
			device=device,
			**kwargs
		)
		

def dummy_train(targets: torch.Tensor):
	targets = to_tensor(targets)[None, ...]
	snn_layer = BellecLIFLayer(
		input_size=targets.shape[-1],
		output_size=targets.shape[-1],
		use_recurrent_connection=True,
		dt=1.0,
		# hh_init="inputs",
		device=targets.device,
	).build()
	output_layer = LILayer(
		input_size=targets.shape[-1],
		output_size=targets.shape[-1],
		device=targets.device,
		# hh_init="inputs",
		use_bias=False,
		dt=1.0,
	).build()
	feedback_weights = torch.randn((snn_layer.recurrent_weights.shape[0], output_layer.forward_weights.shape[0]))
	optimizer = torch.optim.SGD(output_layer.parameters(), lr=0.1, maximize=True, weight_decay=0.0)
	criterion = PVarianceLoss()
	preds = None
	grad_list = []
	p_bar = tqdm(range(1000))
	for i in p_bar:
		eligibility_trace = []
		snn_out_list, snn_hh_list = [targets[:, 0]], [None]
		output_list, hh_list = [targets[:, 0]], [None]
		for t in range(1, targets.shape[1]):
			snn_out, snn_hh = unpack_out_hh(snn_layer(targets[:, t - 1], snn_hh_list[t-1]))
			snn_hh_list.append(snn_hh)
			snn_out_list.append(snn_out)
			out, hh = unpack_out_hh(output_layer(snn_out_list[t], hh_list[t-1]))
			output_list.append(out)
			hh_list.append(hh)
			
			# Compute eligibility trace
			# instantaneous_eligibility_trace = compute_dz_dw_local()
			# out.detach_()
			# eligibility_trace.append(instantaneous_eligibility_trace)
			
		preds = torch.stack(output_list, dim=1)
		output_error = targets[:, 1:] - preds[:, 1:]
		learning_signals = torch.einsum("btf,fn->btn", output_error, feedback_weights)
		# eligibility_trace = torch.stack(eligibility_trace, dim=0)
		v_scaled = (torch.stack([h[0] for h in snn_hh_list[1:]], dim=1) - snn_layer.threshold) / snn_layer.threshold
		post_term = snn_layer.spike_func.pseudo_derivative(v_scaled, snn_layer.threshold, snn_layer.gamma)
		input_spikes = torch.stack(snn_out_list[1:], dim=1)
		z_previous_time = torch.stack(snn_out_list[1:], dim=1)
		z = torch.stack(snn_out_list[1:], dim=1)
		pre_term_w_in = batchwise_temporal_recursive_filter(input_spikes, decay=0.9)
		pre_term_w_rec = batchwise_temporal_recursive_filter(z_previous_time, decay=0.9)
		pre_term_w_out = batchwise_temporal_recursive_filter(z, decay=0.9)
		eligibility_traces_w_in = post_term[:, :, None, :] * pre_term_w_in[:, :, :, None]
		eligibility_traces_w_rec = post_term[:, :, None, :] * pre_term_w_rec[:, :, :, None]
		
		# To define the gradient of the readout error,
		# the eligibility traces are smoothed with the same filter as the readout
		eligibility_traces_convolved_w_in = batchwise_temporal_recursive_filter(eligibility_traces_w_in, decay=0.9)
		eligibility_traces_convolved_w_rec = batchwise_temporal_recursive_filter(eligibility_traces_w_rec, decay=0.9)
		
		# To define the gradient of the regularization error defined on the averaged firing rate,
		# the eligibility traces should be averaged over time
		eligibility_traces_averaged_w_in = eligibility_traces_w_in.mean(dim=(0, 1))
		eligibility_traces_averaged_w_rec = eligibility_traces_w_rec.mean(dim=(0, 1))
		
		# gradients of the main loss with respect to the weights
		dloss_dw_out = torch.einsum("btno->no", output_error[:, :, None, :] * pre_term_w_out[:, :, :, None])
		dloss_dw_in = torch.einsum("btno->no", learning_signals[:, :, None, :] * eligibility_traces_convolved_w_in)
		dloss_dw_rec = torch.einsum("btno->no", learning_signals[:, :, None, :] * eligibility_traces_convolved_w_rec)
		
		# grad = torch.einsum("tno->no", learning_signal[:, np.newaxis] * eligibility_trace)
		snn_layer.forward_weights.grad = dloss_dw_in
		snn_layer.recurrent_weights.grad = dloss_dw_rec
		output_layer.forward_weights.grad = dloss_dw_out
		
		# optimizer.zero_grad()
		loss = criterion(preds, targets)
		# loss.backward()
		optimizer.step()
		
		mean_grad = np.mean(np.abs(to_numpy(snn_layer.forward_weights.grad)))
		grad_list.append(mean_grad)
		p_bar.set_description(f"Loss: {loss.item():.4f}, mean_grad: {mean_grad:.4f}")
	return torch.squeeze(preds), grad_list


if __name__ == '__main__':
	T = 100
	inputs_spikes = torch.rand((1, T, 1)) < 0.1
	# target = torch.stack([torch.cos(torch.linspace(0, 2 * np.pi, 100)) for i in range(100)], dim=-1)
	# target = (target - target.mean(dim=0, keepdim=True)) / target.std(dim=0, keepdim=True)
	# target = (target - target.min(dim=0, keepdim=True)[0]) / (target.max(dim=0, keepdim=True)[0] - target.min(dim=0, keepdim=True)[0])
	target = torch.stack([-torch.exp(torch.linspace(0, 1, T)) for i in range(1)], dim=-1)
	target = (target - target.min(dim=0, keepdim=True)[0]) / (
				target.max(dim=0, keepdim=True)[0] - target.min(dim=0, keepdim=True)[0])
	if target.ndim == 1:
		target = target[:, None]
	predictions, gradients = dummy_train(target)
	
	fig, axes = plt.subplots(1, 2, figsize=(16, 8))
	axes[0].plot(to_numpy(target)[:, 0], 'o', label="target")
	axes[0].plot(to_numpy(predictions), c='k', label="predictions")
	axes[0].legend()
	legend_without_duplicate_labels_(axes[0])
	
	axes[1].plot(gradients, label="mean abs. gradients")
	axes[1].legend()
	plt.show()




from typing import Any, Optional

import numpy as np
import torch
from mlagents_envs.base_env import ActionTuple


class TensorActionTuple:
	def __init__(
			self,
			continuous: Optional[torch.Tensor] = None,
			discrete: Optional[torch.Tensor] = None
	):
		self._continuous = self._cast_entry(continuous)
		self._discrete = self._cast_entry(discrete)
		self.add_empty()

	@property
	def continuous(self) -> Optional[torch.Tensor]:
		return self._continuous

	@property
	def discrete(self) -> Optional[torch.Tensor]:
		return self._discrete

	@staticmethod
	def _cast_entry(entry: Any) -> Optional[torch.Tensor]:
		if entry is None:
			return None
		if isinstance(entry, np.ndarray):
			entry = torch.from_numpy(entry)
		elif isinstance(entry, torch.Tensor):
			pass
		else:
			entry = torch.tensor(entry, dtype=torch.float32)
		return entry if torch.numel(entry) > 0 else None

	def add_empty(self) -> None:
		assert (self._continuous is not None) or (self._discrete is not None), "TensorActionTuple is empty"
		if self._discrete is None:
			self._discrete = torch.empty((self._continuous.shape[0], 0), dtype=torch.long)
		if self._continuous is None:
			self._continuous = torch.empty((self._discrete.shape[0], 0), dtype=torch.float32)

	@staticmethod
	def from_numpy(action_tuple: ActionTuple) -> 'TensorActionTuple':
		return TensorActionTuple(torch.from_numpy(action_tuple.continuous), torch.from_numpy(action_tuple.discrete))

	def to_numpy(self) -> ActionTuple:
		if self._continuous is None:
			continuous_outputs = None
		else:
			continuous_outputs = self._continuous.detach().cpu().numpy()

		if self._discrete is None:
			discrete_outputs = None
		else:
			discrete_outputs = self._discrete.detach().cpu().numpy()
		return ActionTuple(continuous_outputs, discrete_outputs)

	def copy(self) -> 'TensorActionTuple':
		return TensorActionTuple(
			self._continuous.detach().clone() if self._continuous is not None else None,
			self._discrete.detach().clone() if self._discrete is not None else None
		)






from typing import Any

import torch
import enum


class SpikeFuncType(enum.Enum):
	FastSigmoid = 0
	Phi = 1


class SpikeFunction(torch.autograd.Function):
	"""
	Implementation of the spike function. The spike function is a differentiable version of the Heaviside function.
	The Heaviside function is defined as the equation :eq:`heaviside`. The backward pass of this function has to be
	an approximation of the derivative of the Heaviside function.
	
	.. math::
		:label: heaviside
	
		\\begin{equation}
			H(x, thr) = \\left\\{
			\\begin{matrix}
				1 & \\text{ if } x > thr; \\\\
				0 & \\text{ else}.
			\\end{matrix}
			\\right.
		\\end{equation}
		
	"""
	
	@staticmethod
	def forward(
			ctx: torch.autograd.function.FunctionCtx,
			inputs: torch.Tensor,
			threshold: torch.Tensor = torch.tensor(1.0),
			gamma: torch.Tensor = torch.tensor(1.0)
	) -> torch.Tensor:
		"""
		The forward pass of the spike function is the Heaviside function. See the heaviside equation.
		
		:param ctx: The context of the function. It is used to store information for the backward pass. Use the method
			:func:`ctx.save_for_backward` to store information.
		:type ctx: torch.autograd.function.FunctionCtx
		:param inputs: The input tensor.
		:type inputs: torch.Tensor
		:param threshold: The threshold of the spike function.
		:type threshold: torch.Tensor
		:param gamma: The gamma parameter of the spike function. This parameter is used in the backward pass to
			increase the gradient of the spike function. See child classes for more information.
		:type gamma: torch.Tensor
		
		:return: The output of the spike function.
		:rtype: torch.Tensor
		"""
		ctx.save_for_backward(inputs, threshold, gamma)
		out = torch.zeros_like(inputs)
		out[inputs >= threshold] = 1.0
		return out
	
	@staticmethod
	def backward(ctx: torch.autograd.function.FunctionCtx, grad_outputs):
		"""
		In the backward pass we receive a Tensor we need to compute the
		surrogate gradient of the loss with respect to the input.
		Here we use the normalized negative part of a fast sigmoid
		as this was done in Zenke & Ganguli (2018).
		"""
		raise NotImplementedError()

# @staticmethod
# def symbolic(g, inputs: torch._C.Value) -> torch._C.Value:
# 	return g.op("SpikeFunction", inputs, g.op("Constant", value_t=torch.tensor(0, dtype=torch.float)))


class HeavisideSigmoidApprox(SpikeFunction):
	"""
	Implementation of the spike function. The spike function is a differentiable version of the Heaviside function.
	The Heaviside function is defined in the doc of :class:`SpikeFunction`. The backward pass of this function is the
	first derivative of the fast sigmoid function defined in equation :eq:`fast_sigmoid`. The derivative
	is shown in equation :eq:`fast_sigmoid_derivative` used in Zenke & Ganguli (2018).
	
	.. math::
		\\begin{equation}
			S(x) = \\frac{1}{1 + e^{-x}}
		\\end{equation}
		:label: fast_sigmoid
	
	.. math::
		\\begin{equation}
			S'(x) \\approx \\frac{x}{\\left(1 + \\gamma\\vert{x - thr}\\vert\\right)^2}
		\\end{equation}
		:label: fast_sigmoid_derivative
	
	"""
	@staticmethod
	def backward(ctx: torch.autograd.function.FunctionCtx, grad_outputs: torch.Tensor) -> Any:
		"""
		The implementation of the equation :eq:`fast_sigmoid_derivative`.
		
		:param ctx: The context of the function. It is used to retrieve information from the forward pass.
		:type ctx: torch.autograd.function.FunctionCtx
		:param grad_outputs: The gradient of the loss with respect to the output of the forward pass.
		:type grad_outputs: torch.Tensor
		:return: The gradient of the loss with respect to the input of the forward pass.
		"""
		inputs, threshold, scale = ctx.saved_tensors
		grad_outputs_clone = grad_outputs.clone()
		inputs_grad = grad_outputs_clone / (scale * torch.abs(inputs - threshold) + 1.0)**2
		if ctx.needs_input_grad[1]:
			threshold_grad = -inputs_grad.clone()
		else:
			threshold_grad = None
		return inputs_grad, threshold_grad, None


class HeavisidePhiApprox(SpikeFunction):
	"""
	Implementation of the spike function. The spike function is a differentiable version of the Heaviside function.
	The Heaviside function is defined in the doc of :class:`SpikeFunction`. The backward pass of this function is the
	approximation of the heaviside used in :cite:t:`bellec_solution_2020`. This approximation is defined in equation
	:eq:`heaviside_phi_approx`.
	
	.. math::
		\\begin{equation}
			\\psi_j^t = \\frac{\\gamma_\\text{pd}}{v_{\\text{th}}}
			\\text{max}\\left(0, 1 - \\left\\vert\\frac{v_j^t - A_j^t}{v_\\text{th}}\\right\\vert\\right)
		\\end{equation}
		:label: heaviside_phi_approx
	
	.. bibliography::
	
	"""
	epsilon = 1e-5
	
	@staticmethod
	def pseudo_derivative(inputs, threshold, gamma):
		return (gamma / (threshold + HeavisidePhiApprox.epsilon)) * torch.max(
			torch.zeros_like(inputs), 1 - torch.abs((inputs - threshold) / (threshold + HeavisidePhiApprox.epsilon))
		)
	
	@staticmethod
	def backward(ctx: torch.autograd.function.FunctionCtx, grad_outputs):
		"""
		The implementation of the equation :eq:`heaviside_phi_approx`.
		
		:param ctx: The context of the function. It is used to retrieve information from the forward pass.
		:type ctx: torch.autograd.function.FunctionCtx
		:param grad_outputs: The gradient of the loss with respect to the output of the forward pass.
		:type grad_outputs: torch.Tensor
		:return: The gradient of the loss with respect to the input of the forward pass.
		"""
		inputs, threshold, gamma = ctx.saved_tensors
		inputs_grad = grad_outputs.clone() * (gamma / (threshold + HeavisidePhiApprox.epsilon)) * torch.max(
			torch.zeros_like(inputs), 1 - torch.abs((inputs - threshold) / (threshold + HeavisidePhiApprox.epsilon))
		)
		if ctx.needs_input_grad[1]:
			threshold_grad = -inputs_grad.clone()
		else:
			threshold_grad = None
		return inputs_grad, threshold_grad, None


SpikeFuncType2Func = {
	SpikeFuncType.FastSigmoid: HeavisideSigmoidApprox,
	SpikeFuncType.Phi        : HeavisidePhiApprox,
}

if __name__ == '__main__':
	import matplotlib.pyplot as plt
	import numpy as np
	
	th, sc = torch.tensor(1.0), torch.tensor(1.0)
	funcs = {
		"FastSigmoid": HeavisideSigmoidApprox.apply,
		"Bellec"     : HeavisidePhiApprox.apply
	}
	grads = {
		name: [] for name in funcs
	}
	X = torch.tensor(np.linspace(th - 2, th + 2, num=1_000), requires_grad=False)
	Y = SpikeFunction.apply(X, th, sc)
	for name, func in funcs.items():
		for x_i in X:
			x = torch.tensor(x_i.clone().detach(), requires_grad=True)
			y = func(x, th, sc)
			# y.retain_grad()
			y.backward()
			grads[name].append(x.grad.detach().cpu().numpy())
	
	plt.plot(X, Y, label="Heaviside")
	for name in funcs:
		plt.plot(X, grads[name], label=name)
	plt.xlabel("V [mV]")
	plt.legend()
	plt.show()



from typing import Callable, Dict, List, Optional, Union

import torch

from . import Trainer
from ..learning_algorithms.bptt import BPTT
from ..learning_algorithms.learning_algorithm import LearningAlgorithm
from ..transforms.base import ToTensor
from ..metrics import ClassificationMetrics


class ClassificationTrainer(Trainer):
	def __init__(self, *args, **kwargs):
		kwargs.setdefault("y_transform", ToTensor(dtype=torch.long))
		kwargs.setdefault("predict_method", "get_prediction_log_proba")
		super().__init__(*args, **kwargs)
	
	def _set_default_criterion(self, criterion: Optional[torch.nn.Module]) -> torch.nn.Module:
		if criterion is None:
			if isinstance(self.model.output_sizes, dict):
				criterion = {
					k: torch.nn.NLLLoss() for k in self.model.output_sizes
				}
			elif isinstance(self.model.output_sizes, int):
				criterion = torch.nn.NLLLoss()
			else:
				raise ValueError("Unknown criterion type")
		return criterion

	def _set_default_metrics(self, metrics: Optional[List[Callable]]):
		if metrics is None:
			metrics = [ClassificationMetrics(self.model)]
		return metrics
	
	def _maybe_add_learning_algorithm(self, learning_algorithm: Optional[LearningAlgorithm]) -> None:
		if len(self.learning_algorithms) == 0 and learning_algorithm is None:
			learning_algorithm = BPTT(
				optimizer=torch.optim.Adam(self.model.parameters()),
				criterion=self._set_default_criterion(None)
			)
		if learning_algorithm is not None:
			self.callbacks.append(learning_algorithm)






"""
NeuroTorch: A Python library for machine learning and neuroscience.
"""

__author__ = "Jérémie Gince"
__email__ = "gincejeremie@gmail.com"
__copyright__ = "Copyright 2022, Jérémie Gince"
__license__ = "Apache 2.0"
__url__ = "https://github.com/NeuroTorch/NeuroTorch"
__version__ = "v0.0.1-beta0"


from .dimension import (
	Dimension,
	DimensionProperty,
	Size,
	DimensionLike,
	DimensionsLike,
)

from .modules.sequential import (
	Sequential,
)

from .modules.sequential_rnn import (
	SequentialRNN,
)

from .modules.layers import (
	LayerType,
	Linear,
	LILayer,
	LIFLayer,
	SpyLILayer,
	SpyLIFLayer,
	SpyALIFLayer,
	ALIFLayer,
	WilsonCowanLayer,
	WilsonCowanCURBDLayer,
)

from .regularization import (
	RegularizationList,
	L1,
	L2,
)

from .regularization.connectome import (
	DaleLaw,
	DaleLawL2,
)

from .trainers import (
	Trainer,
	ClassificationTrainer,
	RegressionTrainer,
	TrainingState,
)

from .transforms import (
	to_tensor,
	to_numpy,
	IdentityTransform,
	ToDevice,
	ToTensor,
)


from .metrics import (
	losses,
)

from .callbacks import (
	TrainingHistory,
	LoadCheckpointMode,
	CheckpointManager,
)

from . import init

from .learning_algorithms import (
	BPTT,
	TBPTT,
	Eprop,
	RLS,
)

from .visualisation import (
	Visualise,
	VisualiseKMeans,
	VisualisePCA,
	VisualiseUMAP,
)

from .utils import (
	set_seed,
)

import warnings

warnings.filterwarnings("ignore", category=Warning, module="docutils")
warnings.filterwarnings("ignore", category=Warning, module="sphinx")



from copy import deepcopy
from typing import Dict, List, NamedTuple, Optional

# from mlagents_envs.side_channel.environment_parameters_channel import EnvironmentParametersChannel

from .buffers import ReplayBuffer


class CompletionCriteria(NamedTuple):
	"""
	Completion criteria for a lesson.
	"""
	measure: str
	min_lesson_length: int
	threshold: float

	@staticmethod
	def default_criteria() -> 'CompletionCriteria':
		return CompletionCriteria(measure='Rewards', min_lesson_length=1, threshold=0.9)


class Lesson:
	UNPICKLABLE_ATTRIBUTES = ['_teacher', '_channel']

	def __init__(
			self,
			name,
			channel,  #: EnvironmentParametersChannel,
			params: Dict[str, float],
			completion_criteria: CompletionCriteria = CompletionCriteria.default_criteria(),
			teacher=None,
			teacher_strength : Optional[float] = None
	):
		# assert teacher is None or hasattr(teacher, 'buffer'), 'Teacher must have a replay buffer.'
		self.name = name
		self.completion_criteria = completion_criteria
		self._channel = channel
		self.params = params
		self._num_iterations = 0
		self._teacher = teacher
		self.teacher_strength = teacher_strength
		self._result = None

	@property
	def teacher(self):
		return self._teacher

	@teacher.setter
	def teacher(self, teacher):
		self._teacher = teacher

	@property
	def teacher_buffer(self) -> Optional[ReplayBuffer]:
		"""
		Returns the replay buffer for the lesson.
		"""
		if self._teacher is not None:
			buffer = self._teacher.buffer
			assert isinstance(buffer, ReplayBuffer), 'Teacher must have a replay buffer.'
			return self._teacher.buffer
		return None

	@property
	def is_completed(self):
		"""
		Returns True if the lesson is completed, False otherwise.
		"""
		return self._result is not None

	@property
	def channel(self):
		return self._channel

	@channel.setter
	def channel(self, channel):
		self._channel = channel

	def __getstate__(self):
		state = {
			k: v
			for k, v in self.__dict__.items()
			if k not in self.UNPICKLABLE_ATTRIBUTES
		}
		return state

	def __setstate__(self, state):
		self.__dict__.update(state)

	def set_result(self, result):
		self._result = result
		if self._teacher is not None:
			self._teacher.discard_buffer()

	def start(self):
		"""
		Starts the lesson.
		"""
		for key, value in self.params.items():
			self._channel.set_float_parameter(key, value)

	def on_iteration_end(self, metrics: Dict[str, float]) -> bool:
		"""
		Called when an iteration ends.
		"""
		self._num_iterations += 1
		return self.check_completion_criteria(metrics)

	def check_completion_criteria(self, metrics: Dict[str, float]) -> bool:
		"""
		Checks if the lesson is completed.
		"""
		if self.completion_criteria.min_lesson_length > self._num_iterations:
			return False
		result = metrics[self.completion_criteria.measure]
		if result >= self.completion_criteria.threshold:
			self.set_result(result)
		return self.is_completed


class CurriculumEndIterationOutput(NamedTuple):
	"""
	Output of the curriculum when the end of an iteration is reached.
	"""
	messages: Dict[str, str]
	lesson_completed: bool


class Curriculum:
	def __init__(self, name: str = "Curriculum", description: str = "", lessons: List[Lesson] = None):
		self.name = name
		self.description = description
		self._lessons = [] if lessons is None else lessons
		self._current_lesson_idx = 0

	@property
	def is_completed(self) -> bool:
		"""
		Returns True if the curriculum is completed, False otherwise.
		"""
		return self._current_lesson_idx >= len(self._lessons)

	@property
	def current_lesson(self) -> Optional[Lesson]:
		"""
		Returns the current lesson.
		"""
		if self.is_completed:
			return None
		return self._lessons[self._current_lesson_idx]

	@property
	def map_repr(self) -> Dict[str, str]:
		progress = f"{100 * self._current_lesson_idx / (len(self._lessons)-1):.1f} %"
		if self.is_completed:
			return {self.name: f'(lesson: completed) 100%'}
		return {self.name: f'(lesson: {self.current_lesson.name}) {progress}'}

	@property
	def teachers(self):
		return [lesson.teacher for lesson in self._lessons]

	@property
	def channels(self):
		return [lesson.channel for lesson in self._lessons]

	@property
	def lessons(self):
		return self._lessons

	def __getitem__(self, item) -> Lesson:
		if isinstance(item, int):
			return self._lessons[item]
		elif isinstance(item, str):
			return next(filter(lambda l: l.name == item, self._lessons))
		else:
			raise TypeError(f"Invalid type for curriculum indexing: {type(item)}")

	def add_lesson(self, lesson: Lesson):
		self._lessons.append(lesson)

	def __str__(self):
		map_repr = self.map_repr
		return f'{self.name} {map_repr[self.name]}'

	def __repr__(self):
		return str(self)

	@property
	def teacher_buffer(self) -> Optional[ReplayBuffer]:
		"""
		Returns the current teacher buffer.
		"""
		if self.is_completed:
			return None
		return self._lessons[self._current_lesson_idx].teacher_buffer

	def on_iteration_start(self):
		"""
		Called when an iteration starts.
		"""
		if not self.is_completed:
			self.current_lesson.start()

	def on_iteration_end(self, metrics: Dict[str, float]) -> CurriculumEndIterationOutput:
		"""
		Called when an iteration ends.
		"""
		lesson_is_completed = False
		if not self.is_completed:
			self.current_lesson.on_iteration_end(metrics)
			if self.current_lesson.is_completed:
				self._current_lesson_idx += 1
				lesson_is_completed = True
		return CurriculumEndIterationOutput(messages=self.map_repr, lesson_completed=lesson_is_completed)

	def update_teachers(self, teachers: List):
		assert len(teachers) == len(self._lessons), 'Number of teachers must match number of lessons.'
		for lesson, teacher in zip(self._lessons, teachers):
			lesson.teacher = teacher

	def update_channels(self, channels: List):
		assert len(channels) == len(self._lessons), 'Number of channels must match number of lessons.'
		for lesson, channel in zip(self._lessons, channels):
			lesson._channel = channel

	def update_teachers_and_channels(self, other: 'Curriculum'):
		self.update_teachers(other.teachers)
		self.update_channels(other.channels)





from typing import Union, Dict, Any, Tuple, Type, Optional

import torch

from .spikes_decoders import MeanConv
from .spikes_encoders import SpikesEncoder
from ..modules.layers import SpyLIFLayer, LIFLayer, ALIFLayer
from ..modules.base import BaseModel


class SpikesAutoEncoder(BaseModel):
	
	@staticmethod
	def _default_encoder_type() -> Type[Union[LIFLayer, SpyLIFLayer, ALIFLayer]]:
		return SpyLIFLayer
	
	def __init__(
			self,
			n_units: int,
			n_encoder_steps: int,
			encoder_type: Optional[Type[Union[LIFLayer, SpyLIFLayer, ALIFLayer]]] = None,
			spikes_encoder: Optional[torch.nn.Module] = None,
			spikes_decoder: Optional[torch.nn.Module] = None,
			**kwargs
	):
		if encoder_type is not None and spikes_encoder is not None:
			raise ValueError(
				"If encoder_type is provided, spikes_encoder must be None and vice versa."
			)
		kwargs.setdefault("name", "SpikesAutoEncoder")
		super().__init__(n_units, n_units, **kwargs)
		if encoder_type is None:
			encoder_type = self._default_encoder_type()
		self.encoder_type = encoder_type
		self.n_units = n_units
		self.n_encoder_steps = n_encoder_steps
		if spikes_encoder is None:
			spikes_encoder = self._create_default_encoder()
		self.spikes_encoder = spikes_encoder
		if spikes_decoder is None:
			spikes_decoder = self._create_default_decoder()
		self.spikes_decoder = spikes_decoder
	
	@BaseModel.device.setter
	def device(self, device: torch.device):
		self.spikes_encoder.to(device)
		self.spikes_decoder.to(device)
		BaseModel.device.fset(self, device)
		
	def _create_default_encoder(self):
		return SpikesEncoder(
			n_steps=self.n_encoder_steps,
			n_units=self.n_units,
			spikes_layer_type=self.encoder_type,
			device=self.device,
			dt=self.kwargs.pop("dt", 1e-3),
			**self.kwargs,
		)
	
	def _create_default_decoder(self):
		return MeanConv(
			self.n_encoder_steps,
			alpha=2.0,
			learn_alpha=True,
			learn_kernel=True,
			activation=torch.nn.Hardtanh(
				min_val=self.kwargs.get("hard_tanh.min_val", 0.0),
				max_val=self.kwargs.get("hard_tanh.max_val", 1.0)
			),
		)
	
	def encode(self, x: torch.Tensor) -> torch.Tensor:
		return self.spikes_encoder(x)
	
	def decode(self, x: torch.Tensor):
		return self.spikes_decoder(x)
	
	def forward(self, x: torch.Tensor, **kwargs):
		x = self._to_device_transform(x)
		return self.decode(self.encode(x))
	
	def get_prediction_trace(
			self, inputs: Union[Dict[str, Any], torch.Tensor],
			**kwargs
	) -> Union[Dict[str, torch.Tensor], torch.Tensor]:
		raise NotImplementedError()
	
	def get_raw_prediction(
			self,
			inputs: torch.Tensor,
			*args,
			**kwargs
	) -> Union[Tuple[Any, Any, Any], Tuple[Any, Any], Any]:
		return self.forward(inputs)
	
	def get_and_reset_regularization_loss(self) -> torch.Tensor:
		"""
		Get the regularization loss as a sum of all the regularization losses of the layers. Then reset the
		regularization losses.

		:return: the regularization loss.
		"""
		regularization_loss = torch.tensor(0.0, dtype=torch.float32, device=self.device)
		for layer in [self.spikes_encoder, self.spikes_decoder]:
			if hasattr(layer, "get_and_reset_regularization_loss") and callable(
					layer.get_and_reset_regularization_loss
					):
				regularization_loss += layer.get_and_reset_regularization_loss()
		return regularization_loss












from .base import (
	to_tensor,
	to_numpy,
	LinearRateToSpikes,
	ConstantValuesTransform,
	IdentityTransform,
	ToDevice,
	ToTensor,
)

# from .spikes_encoders import (
# 	SpikesEncoder,
# 	LIFEncoder,
# 	SpyLIFLayer,
# 	ALIFEncoder,
# )






import numbers
from typing import Any, Optional

import numpy as np
import torch


def to_tensor(x: Any, dtype=torch.float32):
	if isinstance(x, np.ndarray):
		return torch.from_numpy(x).type(dtype)
	elif isinstance(x, torch.Tensor):
		return x.type(dtype)
	elif not isinstance(x, torch.Tensor):
		return torch.tensor(x, dtype=dtype)
	raise ValueError(f"Unsupported type {type(x)}")


def to_numpy(x: Any, dtype=np.float32):
	if isinstance(x, np.ndarray):
		return np.asarray(x, dtype=dtype)
	elif isinstance(x, torch.Tensor):
		return x.detach().cpu().numpy()
	elif isinstance(x, numbers.Number):
		return x
	elif isinstance(x, dict):
		return {k: to_numpy(v, dtype=dtype) for k, v in x.items()}
	elif not isinstance(x, torch.Tensor):
		return np.asarray(x, dtype=dtype)
	raise ValueError(f"Unsupported type {type(x)}")


class ToDevice(torch.nn.Module):
	def __init__(self, device: torch.device, non_blocking: bool = True):
		super().__init__()
		self.device = device
		self.non_blocking = non_blocking

	def forward(self, x: torch.Tensor):
		if x is None:
			return x
		if not isinstance(x, torch.Tensor):
			if isinstance(x, dict):
				return {k: self.forward(v) for k, v in x.items()}
			elif isinstance(x, list):
				return [self.forward(v) for v in x]
			elif isinstance(x, tuple):
				return tuple(self.forward(v) for v in x)
			else:
				return x
		return x.to(self.device, non_blocking=self.non_blocking)
	
	def __repr__(self):
		return f"ToDevice({self.device}, async={self.non_blocking})"


class ToTensor(torch.nn.Module):
	def __init__(self, dtype=torch.float32, device: Optional[torch.device] = None):
		super().__init__()
		self.dtype = dtype
		self.device = device
		if self.device is None:
			self.to_device = None
		else:
			self.to_device = ToDevice(self.device)
	
	def forward(self, x: Any) -> torch.Tensor:
		x = to_tensor(x, self.dtype)
		if self.to_device is not None:
			x = self.to_device(x)
		return x


class IdentityTransform(torch.nn.Module):
	def __init__(self):
		super().__init__()
	
	def forward(self, x: Any):
		return x


class LinearRateToSpikes(torch.nn.Module):
	def __init__(
			self,
			n_steps: int,
			*,
			data_min: float = 0.0,
			data_max: float = 1.0,
			epsilon: float = 1e-6,
	):
		super().__init__()
		self.n_steps = n_steps
		self.data_min = to_tensor(data_min)
		self.data_max = to_tensor(data_max)
		self.epsilon = epsilon
	
	def __repr__(self):
		return f"{self.__class__.__name__}(n_steps={self.n_steps})"

	def firing_periods_to_spikes(self, firing_periods: np.ndarray) -> np.ndarray:
		firing_periods = np.floor(firing_periods).astype(int)
		ones_mask = firing_periods == 1
		firing_periods[firing_periods > self.n_steps] = self.n_steps
		firing_periods[firing_periods < 1] = 1
		time_dim = np.expand_dims(np.arange(1, self.n_steps+1), axis=tuple(np.arange(firing_periods.ndim)+1))
		spikes = (time_dim % firing_periods) == 0
		spikes[0, ones_mask] = 0
		return spikes.astype(float)

	def forward(self, x):
		x = to_tensor(x)
		device = x.device
		squeeze_flag = False
		if x.dim() < 1:
			x = x.unsqueeze(0)
			squeeze_flag = True
		self.data_min = torch.minimum(x, self.data_min)
		self.data_max = torch.maximum(x, self.data_max)
		x = (x - self.data_min) / (self.data_max - self.data_min + self.epsilon)
		periods = torch.floor((1 - x) * self.n_steps).type(torch.long)
		spikes = to_tensor(self.firing_periods_to_spikes(periods.cpu().numpy()))
		if squeeze_flag:
			spikes = spikes.squeeze()
		return spikes.to(device)


class ConstantValuesTransform(torch.nn.Module):
	def __init__(self, n_steps: int, batch_wise: bool = True):
		super().__init__()
		self.n_steps = n_steps
		self.batch_wise = batch_wise
	
	def __repr__(self):
		return f"{self.__class__.__name__}(n_steps={self.n_steps})"

	def forward(self, x: Any):
		x = to_tensor(x)
		if self.batch_wise:
			return x.repeat(1, self.n_steps, 1)
		return x.repeat(self.n_steps, 1)




import torch
import torch.nn as nn

from ..transforms.base import to_tensor


class RMSELoss(torch.nn.Module):
	"""
	Class used to compute the RMSE loss.
	
	:math:`\\text{RMSE}(x, y) = \\sqrt{\\frac{1}{n}\\sum_{i=1}^n (x_i - y_i)^2}`
	
	:Attributes:
		- **criterion** (nn.MSELoss): The MSE loss.
	
	"""
	def __init__(self):
		"""
		Constructor for the RMSELoss class.
		"""
		super(RMSELoss, self).__init__()
		self.criterion = nn.MSELoss()

	def forward(self, x, y):
		"""
		Calculate the RMSE loss.
		
		:param x: The first input.
		:param y: The second input.
		
		:return: The RMSE loss.
		"""
		loss = self.criterion(x, y)
		loss = torch.pow(loss + 1e-8, 0.5)
		return loss


class PVarianceLoss(torch.nn.Module):
	"""
	Class used to compute the P-Variance loss.
	
	:math:`\\text{P-Variance}(x, y) = 1 - \\frac{\\text{MSE}(x, y)}{\\text{Var}(y)}`
	
	:Attributes:
		- :attr:`criterion` (nn.MSELoss): The MSE loss.
		- :attr:`negative` (bool): Whether to return the negative P-Variance loss.
		- :attr:`reduction` (str): The reduction method to use. If 'mean', the output will be averaged. If 'feature', the
			output will be the shape of the last dimension of the input. If 'none', the output will be the same shape as
			the input.
	
	"""
	def __init__(self, negative: bool = False, reduction: str = 'mean', **kwargs):
		"""
		Constructor for the PVarianceLoss class.
		
		:param negative: Whether to return the negative P-Variance loss.
		:type negative: bool
		:param reduction: The reduction method to use. If 'mean', the output will be averaged. If 'feature', the output
			will be the shape of the last dimension of the input. If 'none', the output will be the same shape as the
			input. Defaults to 'mean'.
		:type reduction: str
		:keyword arguments : epsilon: The epsilon value to use to prevent division by zero. Defaults to 1e-5.
		"""
		super(PVarianceLoss, self).__init__()
		assert reduction in ['mean', 'feature', 'none'], 'Reduction must be one of "mean", "feature", or "none".'
		self.reduction = reduction
		mse_reduction = 'mean' if reduction == 'mean' else 'none'
		self.criterion = nn.MSELoss(
			reduction=mse_reduction
		)
		self.negative = negative
		self.epsilon = kwargs.get("epsilon", 1e-5)

	def forward(self, x, y):
		"""
		Calculate the P-Variance loss.
		
		:param x: The first input.
		:param y: The second input.
		
		:return: The P-Variance loss.
		"""
		x, y = to_tensor(x), to_tensor(y)
		if self.reduction == 'feature':
			x_reshape, y_reshape = x.reshape(-1, x.shape[-1]), y.reshape(-1, y.shape[-1])
		else:
			x_reshape, y_reshape = x, y
		mse_loss = self.criterion(x_reshape, y_reshape)
		if self.reduction == 'feature':
			mse_loss = mse_loss.mean(dim=0)
			var = y_reshape.var(dim=0)
		else:
			var = y_reshape.var()
		loss = 1 - mse_loss / (var + self.epsilon)
		if self.negative:
			loss = -loss
		return loss
	
	def mean_std_over_batch(self, x, y):
		"""
		Calculate the mean and standard deviation of the P-Variance loss over the batch.
		
		:param x: The first input.
		:param y: The second input.
		
		:return: The mean and standard deviation of the P-Variance loss over the batch.
		"""
		x, y = to_tensor(x), to_tensor(y)
		x_reshape, y_reshape = x.reshape(x.shape[0], -1), y.reshape(y.shape[0], -1)
		mse_loss = torch.mean((x_reshape - y_reshape)**2, dim=-1)
		var = y_reshape.var(dim=-1)
		loss = 1 - mse_loss / (var + self.epsilon)
		if self.negative:
			loss = -loss
		return loss.mean(), loss.std()



import os
from typing import Dict, List, Tuple, Optional

import numpy as np
import matplotlib.pyplot as plt

from ..transforms.base import to_numpy
from ..callbacks.base_callback import BaseCallback


class TrainingHistory(BaseCallback):
	"""
	This class is used to store some metrics over the training process.
	
	:Attributes:
		- **default_value** (float): The default value to use to equalize the lengths of the container's items.
	"""
	DEFAULT_PRIORITY = BaseCallback.DEFAULT_HIGH_PRIORITY

	@staticmethod
	def _set_default_plot_kwargs(kwargs: dict):
		kwargs.setdefault('fontsize', 18)
		kwargs.setdefault('linewidth', 4)
		kwargs.setdefault('figsize', (16, 12))
		kwargs.setdefault('dpi', 300)
		return kwargs

	@staticmethod
	def _remove_prefix_from_metrics(metrics: List[str]):
		return [metric.split('_')[-1] for metric in metrics]

	def __init__(
			self,
			container: Dict[str, List[float]] = None,
			default_value=np.NAN,
			**kwargs
	):
		"""
		Initialize the container with the given container.
		
		:param container: The container to initialize the container with.
		:type container: Dict[str, List[float]]
		:param default_value: The default value to use to equalize the lengths of the container's items.
		:type default_value: float
		:param kwargs: The keyword arguments to pass to the BaseCallback.
		"""
		super().__init__(**kwargs)
		self._container = {}
		self.default_value = default_value
		self._length = 0
		if container is not None:
			self.concat(container)

	def __getitem__(self, key):
		self._add_key(key)
		return self._container[key]

	def __setitem__(self, key, value: list):
		self._add_key(key)
		assert len(value) == len(self), "Length of value must be equal to length of container"
		self._container[key] = list(value)

	def __contains__(self, item):
		return item in self._container

	def __iter__(self):
		return iter(self._container)

	def __len__(self):
		return self._length

	def _add_key(self, key):
		if key not in self._container:
			self._container[key] = [self.default_value] * len(self)

	def _increase_to_size(self, to_size: int):
		"""
		Increase the size of the container items to the given size.
		
		:param to_size: The size to increase the container to.
		:type to_size: int
		
		:return: None
		"""
		if len(self) > to_size:
			raise ValueError("Cannot increase size of container to smaller size")
		if len(self) == to_size:
			return
		for key, values in self.items():
			self[key].extend([self.default_value] * (to_size - len(self)))
		self._length = to_size

	def keys(self):
		return self._container.keys()

	def items(self):
		return self._container.items()

	def concat(self, other):
		self.insert(len(self), other)

	def insert(self, index: int, other):
		"""
		Increase the size of the container items to the given index and insert the given other into the container.
		
		:param index: The index to insert the other at.
		:type index: int
		:param other: The other to insert.
		:type other: Dict[str, float]
		
		:return: None
		"""
		if index >= len(self):
			self._increase_to_size(index + 1)
		for key, values in other.items():
			if isinstance(values, list):
				raise NotImplementedError()
			else:
				self[key][index] = values

	def append(self, key, value):
		"""
		Add the given value to the given key. Increase the size of the container by one.
		
		:param key: The key to add the value to.
		:type key: str
		:param value: The value to add.
		:type value: float
		
		:return: None
		"""
		self.insert(len(self), {key: value})

	def min(self, key=None, default: float = np.inf) -> float:
		"""
		Get the minimum value of the given key.
		
		:param key: The key to get the minimum value of. If None, the first key is used.
		:type key: str
		:param default: The default value to return if the key is not in the container.
		:type default: float
		
		:return: The minimum value of the given key.
		:rtype: float
		"""
		if key is None:
			key = list(self.keys())[0]
		if key in self:
			return np.nanmin(self[key])
		return default

	def min_item(self, key=None) -> Dict[str, float]:
		"""
		Get all the metrics of the iteration with the minimum value of the given key.
		
		:param key: The key to get the minimum value of. If None, the first key is used.
		:type key: str
		
		:return: All the metrics of the iteration with the minimum value of the given key.
		:rtype: Dict[str, float]
		
		:raises ValueError: If the key is not in the container.
		"""
		if key is None:
			key = list(self.keys())[0]
		if key in self:
			argmin = np.argmin(self[key])
			return {k: v[argmin] for k, v in self.items()}
		raise ValueError("key not in container")

	def max(self, key=None, default=-np.inf):
		"""
		Get the maximum value of the given key.
		
		:param key: The key to get the maximum value of. If None, the first key is used.
		:type key: str
		:param default: The default value to return if the key is not in the container.
		:type default: float
		
		:return: The maximum value of the given key.
		:rtype: float
		"""
		if key is None:
			key = list(self.keys())[0]
		if key in self:
			return np.nanmax(self[key])
		return default

	def max_item(self, key=None):
		"""
		Get all the metrics of the iteration with the maximum value of the given key.

		:param key: The key to get the maximum value of. If None, the first key is used.
		:type key: str

		:return: All the metrics of the iteration with the maximum value of the given key.
		:rtype: Dict[str, float]

		:raises ValueError: If the key is not in the container.
		"""
		if key is None:
			key = list(self.keys())[0]
		if key in self:
			argmax = np.argmax(self[key])
			return {k: v[argmax] for k, v in self.items()}
		raise ValueError("key not in container")
	
	def get_item_at(self, idx: int = -1):
		"""
		Get all the metrics of the iteration at the given index.
		
		:param idx: The index to get the metrics of.
		:type idx: int
		
		:return: All the metrics of the iteration at the given index.
		:rtype: Dict[str, float]
		
		:raises ValueError: If the index is out of bounds.
		"""
		if idx >= len(self):
			raise ValueError("Index out of bounds")
		return {k: v[idx] for k, v in self.items()}

	def create_plot(self, **kwargs) -> Tuple[plt.Figure, Dict[str, plt.Axes], Dict[str, plt.Line2D]]:
		"""
		Create a plot of the metrics in the container.
		
		:param kwargs: Keyword arguments.
		
		:keyword Tuple[float, float] figsize: The size of the figure.
		:keyword int linewidth: The width of the lines.
		
		:return: The figure, axes and lines of the plot.
		:rtype: Tuple[plt.Figure, Dict[str, plt.Axes], Dict[str, plt.Line2D]]
		"""
		kwargs = self._set_default_plot_kwargs(kwargs)
		keys_lower_to_given = {k.lower(): k for k in self.keys()}
		keys_lower = [key.lower() for key in self.keys()]
		loss_metrics = [k for k in keys_lower if 'loss' in k]
		keys_lower = list(set(keys_lower) - set(loss_metrics))
		val_metrics = [k for k in keys_lower if 'val' in k]
		train_metrics = [k for k in keys_lower if 'train' in k]
		test_metrics = [k for k in keys_lower if 'test' in k]
		n_set_metrics = max(len(val_metrics), len(train_metrics), len(test_metrics))
		max_set_metrics_container = [c for c in [val_metrics, train_metrics, test_metrics] if len(c) == n_set_metrics][0]
		other_metrics = list(set(keys_lower) - set(val_metrics) - set(train_metrics) - set(test_metrics))
		n_graphs = 1 + n_set_metrics + len(other_metrics)
		n_cols = int(np.sqrt(n_graphs))
		n_rows = int(np.ceil(n_graphs / n_cols))
		axes_dict, lines = {}, {}
		fig, axes = plt.subplots(nrows=n_rows, ncols=n_cols, figsize=kwargs["figsize"], sharex='all')
		axes = np.ravel(axes)
		for i, ax in enumerate(axes):
			if i >= n_graphs:
				ax.axis('off')
				continue
			if i == 0:
				for k in loss_metrics:
					key = keys_lower_to_given[k]
					lines[key] = ax.plot(self[key], label=key, linewidth=kwargs['linewidth'])[0]
				axes_dict['losses'] = ax
				ax.set_ylabel("Loss [-]", fontsize=kwargs["fontsize"])
				ax.set_xlabel("Iterations [-]", fontsize=kwargs["fontsize"])
				ax.legend(fontsize=kwargs["fontsize"])
			elif 0 < i <= n_set_metrics:
				metric_basename = '_'.join(max_set_metrics_container[i-1].split('_')[1:])
				for prefix in ['val', 'train', 'test']:
					k = prefix + '_' + metric_basename
					if k in keys_lower_to_given:
						key = keys_lower_to_given[k]
						if key in self:
							lines[key] = ax.plot(self[key], label=key, linewidth=kwargs['linewidth'])[0]
							axes_dict[key] = ax
				ax.set_xlabel("Iterations [-]", fontsize=kwargs["fontsize"])
				ax.legend(fontsize=kwargs["fontsize"])
			else:
				k = other_metrics[i - 1 - n_set_metrics]
				key = keys_lower_to_given[k]
				lines[key] = ax.plot(self[key], label=key, linewidth=kwargs['linewidth'])[0]
				axes_dict[key] = ax
				ax.set_xlabel("Iterations [-]", fontsize=kwargs["fontsize"])
				ax.legend(fontsize=kwargs["fontsize"])
		return fig, axes_dict, lines

	def plot(
			self,
			save_path: Optional[str] = None,
			show: bool = False,
			**kwargs
	) -> Tuple[plt.Figure, Dict[str, plt.Axes], Dict[str, plt.Line2D]]:
		"""
		Plot the metrics in the container.
		
		:param save_path: The path to save the plot to. If None, the plot is not saved.
		:type save_path: Optional[str]
		
		:param show: Whether to show the plot.
		:type show: bool
		
		:param kwargs: Keyword arguments.
		
		:keyword Tuple[float, float] figsize: The size of the figure.
		:keyword int linewidth: The width of the lines.
		:keyword int dpi: The resolution of the figure.
		:keyword bool block: Whether to block execution until the plot is closed.
		:keyword bool close: Whether to close the plot at the end.
		
		:return: The figure, axes and lines of the plot.
		:rtype: Tuple[plt.Figure, Dict[str, plt.Axes], Dict[str, plt.Line2D]]
		"""
		kwargs = self._set_default_plot_kwargs(kwargs)
		plt.close('all')
		fig, axes, lines = self.create_plot(**kwargs)
		if save_path is not None:
			os.makedirs(os.path.dirname(save_path), exist_ok=True)
			fig.savefig(save_path, dpi=kwargs["dpi"])
		if show:
			plt.show(block=kwargs.get('block', True))
		if kwargs.get('close', True):
			plt.close(fig)
		return fig, axes, lines

	def update_fig(
			self,
			fig: plt.Figure,
			axes: Dict[str, plt.Axes],
			lines: Dict[str, plt.Line2D],
			**kwargs
	) -> Tuple[plt.Figure, Dict[str, plt.Axes], Dict[str, plt.Line2D]]:
		"""
		Update the plot of the metrics in the container.
		
		:param fig: The figure to update.
		:type fig: plt.Figure
		:param axes: The axes to update.
		:type axes: Dict[str, plt.Axes]
		:param lines: The lines to update.
		:type lines: Dict[str, plt.Line2D]
		:param kwargs: Keyword arguments.
		
		:return: The figure, axes and lines of the plot.
		:rtype: Tuple[plt.Figure, Dict[str, plt.Axes], Dict[str, plt.Line2D]]
		"""
		kwargs = self._set_default_plot_kwargs(kwargs)
		for k in lines:
			lines[k].set_data(range(len(self[k])), self[k])
		for k in axes:
			axes[k].relim()
			axes[k].autoscale_view()
		fig.canvas.draw()
		fig.canvas.flush_events()
		return fig, axes, lines

	def on_iteration_end(self, trainer, **kwargs):
		"""
		Insert the current metrics into the container.
		
		:param trainer: The trainer.
		:type trainer: Trainer
		
		:return: None
		"""
		self.insert(
			trainer.current_training_state.iteration,
			{k: to_numpy(v) for k, v in trainer.current_training_state.itr_metrics.items()}
		)
		
	def extra_repr(self):
		return f", n={len(self)}, metrics={list(self.keys())}"










from typing import Optional, Any, Dict, Union, List, Callable

import torch
from torch.utils.data import DataLoader
from tqdm.auto import tqdm

from ..modules import BaseModel


class BaseMetrics:
	"""
	This base class is used to compute metrics for a model.
	
	:Attributes:
		- **model** (BaseModel): The model to evaluate.
		- **metrics_names** (List[str]): The metrics to compute.
		- **device** (torch.device): The device in which the metrics will be computed.
	"""
	METRICS_NAMES_SEP = ' '
	
	def __init__(
			self,
			model: BaseModel,
			metrics_names: Any = "all",
			device: Optional[torch.device] = None
	):
		"""
		Constructor for the BaseMetrics class.

		:param model: the model to evaluate.
		:type model: BaseModel
		:param metrics_names: the metrics to compute.
		:type metrics_names: Any
		:param device: The device in which the metrics will be computed.
		:type device: Optional[torch.device]
		"""
		self.model = model
		self.metrics_names = self._format_metrics_names_(metrics_names)
		self._check_metrics_names_(self.metrics_names)
		self.device = device
	
	@property
	def metrics_functions(self) -> Dict[str, Callable]:
		"""
		Get the metrics functions to compute.
		
		:return: The metrics functions to compute.
		:rtype: Dict[str, Callable]
		"""
		all_metrics_names_to_func = self.get_unwrap_all_metrics_names_to_func()
		return {
			metric_name: all_metrics_names_to_func[metric_name]
			for metric_name in self.metrics_names
		}
	
	@classmethod
	def _format_metrics_names_(cls, metrics: Any) -> List[str]:
		if isinstance(metrics, str) and metrics.lower() == "all":
			return cls.get_unique_metrics_names()
		if isinstance(metrics, str):
			metrics = metrics.split(cls.METRICS_NAMES_SEP)
		elif isinstance(metrics, list):
			metrics = [metric.strip() for metric in metrics]
		else:
			raise ValueError("metrics must be a string or a list of strings")
		return metrics
	
	@classmethod
	def _check_metrics_names_(cls, metrics: List[str]) -> None:
		all_metrics_names = cls.get_all_metrics_names()
		assert all([metric in all_metrics_names for metric in metrics]), \
			f"metrics must be in {all_metrics_names}"
	
	@staticmethod
	def get_all_metrics_names_to_func() -> Dict[str, Callable]:
		"""
		Get all the metrics names and their associated functions.
		
		:raises NotImplementedError: This method must be implemented in the child class.
		"""
		raise NotImplementedError()
	
	@classmethod
	def get_unwrap_all_metrics_names_to_func(cls) -> Dict[str, Callable]:
		"""
		Get all the metrics names and their associated functions. The metrics names are unwrapped.
		
		:return: The metrics names and their associated functions.
		:rtype: Dict[str, Callable]
		"""
		return {
			metric_name: metric_func
			for metric_names, metric_func in cls.get_all_metrics_names_to_func().items()
			for metric_name in metric_names.split(cls.METRICS_NAMES_SEP)
		}
	
	@classmethod
	def get_all_metrics_names(cls) -> List[str]:
		"""
		Get all the metrics names.
		
		:return: The metrics names.
		:rtype: List[str]
		"""
		all_metrics_names = []
		for metric_names, _ in cls.get_all_metrics_names_to_func().items():
			all_metrics_names.extend(metric_names.split(cls.METRICS_NAMES_SEP))
		return all_metrics_names
	
	@classmethod
	def get_unique_metrics_names(cls) -> List[str]:
		"""
		Get all the unique metrics names.
		
		:return: The unique metrics names.
		:rtype: List[str]
		"""
		all_metrics_names = []
		for metric_names, _ in cls.get_all_metrics_names_to_func().items():
			all_metrics_names.append(metric_names.split(cls.METRICS_NAMES_SEP)[0])
		return all_metrics_names
	
	def __call__(
			self,
			data_loader: DataLoader,
			verbose: Union[bool, int] = False
	) -> Dict[str, Any]:
		"""
		Compute the metrics for the given data_loader.
		
		:param data_loader: The data loader to use to compute the metrics.
		:type data_loader: DataLoader
		:param verbose: 0: no progress bar, 1: single progress bar, 2: progress bar for each metrics
						True: verbose is set to 1, False: verbose is set to 0.
		:type verbose: Union[bool, int]
		
		:return: The metrics computed.
		:rtype: Dict[str, Any]
		"""
		if isinstance(verbose, bool):
			verbose = 1 if verbose else 0
		assert verbose in [0, 1, 2], "verbose must be 0, 1 or 2"
		output = {}
		p_bar = tqdm(
			enumerate(self.metrics_functions.items()),
			total=len(self.metrics_functions),
			disable=verbose == 0,
			unit="metric",
			position=0,
		)
		self.model.eval()
		for i, (metric_name, metric_func) in p_bar:
			output[metric_name] = metric_func(
				model=self.model,
				dataloader=data_loader,
				device=self.device,
				verbose=verbose == 2,
				desc=metric_name,
				p_bar_position=i + 1
			)
		return output

















from typing import List, Optional, Sequence

import torch

from ..callbacks.base_callback import BaseCallback


class LearningAlgorithm(BaseCallback):
	DEFAULT_PRIORITY = BaseCallback.DEFAULT_MEDIUM_PRIORITY
	
	def __init__(
			self,
			*,
			params: Optional[Sequence[torch.nn.Parameter]] = None,
			**kwargs
	):
		"""
		Constructor for LearningAlgorithm class.
		
		:param params: The parameters to optimize. If None, the parameters of the model's trainer will be used.
		:type params: Optional[Sequence[torch.nn.Parameter]]
		"""
		super().__init__(**kwargs)
		if params is None:
			params = []
		else:
			params = list(params)
		self.params: List[torch.nn.Parameter] = params






import pickle
from copy import deepcopy
from typing import Any, Iterable, List, NamedTuple, Optional

import numpy as np
import torch
from queue import PriorityQueue

from .wrappers import TensorActionTuple


class Experience:
	"""
	An experience contains the data of one Agent transition.
	- Observation
	- Action
	- Reward
	- Terminal flag
	- Next Observation
	"""
	def __init__(
			self,
			obs: Any,
			action: TensorActionTuple,
			reward: float,
			terminal: bool,
			next_obs: Any,
			discounted_reward: Optional[float] = None,
			advantage: Optional[float] = None,
	):
		self.obs = obs
		self.action = action
		self.reward = reward
		self.terminal = terminal
		self.next_obs = next_obs
		self._discounted_reward = discounted_reward
		self._advantage = advantage

	@property
	def discounted_reward(self) -> float:
		if self._discounted_reward is None:
			return self.reward
		return self._discounted_reward

	@discounted_reward.setter
	def discounted_reward(self, value: float):
		self._discounted_reward = value

	@property
	def advantage(self) -> float:
		if self._advantage is None:
			return self.discounted_reward
		return self._advantage

	@advantage.setter
	def advantage(self, value: float):
		self._advantage = value


class BatchExperience:
	def __init__(
			self,
			batch: List[Experience],
			device='cpu'
	):
		"""
		An object that contains a batch of experiences as tensors.

		:param batch: A list of Experience objects.
		:param device: The device to use for the tensors.
		"""
		self._device = device
		self._nb_obs = len(batch[0].obs)

		self.obs: List[torch.Tensor] = self._make_obs_batch(batch)
		self.rewards: torch.Tensor = self._make_rewards_batch(batch)
		self.terminals: torch.Tensor = self._make_terminals_batch(batch)
		self.continuous_actions = torch.stack([ex.action.continuous.to(device) for ex in batch])
		self.discrete_actions = torch.stack([ex.action.discrete.to(device) for ex in batch])
		self.next_obs: List[torch.Tensor] = self._make_next_obs_batch(batch)
		self.discounted_rewards: torch.Tensor = self._make_discounted_rewards_batch(batch)
		self.advantages: torch.Tensor = self._make_advantages_batch(batch)

	@property
	def device(self):
		return self._device

	@device.setter
	def device(self, value):
		self._device = value
		self.obs = [obs.to(value) for obs in self.obs]
		self.next_obs = [next_obs.to(value) for next_obs in self.next_obs]
		self.rewards = self.rewards.to(value)
		self.terminals = self.terminals.to(value)
		self.continuous_actions = self.continuous_actions.to(value)
		self.discrete_actions = self.discrete_actions.to(value)
		self.discounted_rewards = self.discounted_rewards.to(value)
		self.advantages = self.advantages.to(value)

	def __len__(self):
		return self.rewards.shape[0]

	def _make_obs_batch(self, batch: List[Experience]) -> List[torch.Tensor]:
		return [
			torch.from_numpy(np.stack([ex.obs[i] for ex in batch])).to(self.device)
			for i in range(self._nb_obs)
		]

	def _make_next_obs_batch(self, batch: List[Experience]) -> List[torch.Tensor]:
		return [
			torch.from_numpy(np.stack([ex.next_obs[i] for ex in batch])).to(self.device)
			for i in range(self._nb_obs)
		]

	def _make_rewards_batch(self, batch: List[Experience]) -> torch.Tensor:
		return torch.from_numpy(
			np.array([ex.reward for ex in batch], dtype=np.float32).reshape(-1, 1)
		).to(self.device)

	def _make_terminals_batch(self, batch: List[Experience]) -> torch.Tensor:
		return torch.from_numpy(
			np.array([ex.terminal for ex in batch], dtype=np.float32).reshape(-1, 1)
		).to(self.device)

	def _make_discounted_rewards_batch(self, batch: List[Experience]) -> torch.Tensor:
		return torch.from_numpy(
			np.array([ex.discounted_reward for ex in batch], dtype=np.float32).reshape(-1, 1)
		).to(self.device)

	def _make_advantages_batch(self, batch: List[Experience]) -> torch.Tensor:
		return torch.from_numpy(
			np.array([ex.advantage for ex in batch], dtype=np.float32).reshape(-1, 1)
		).to(self.device)


class Trajectory:
	"""
	A trajectory is a list of experiences.
	"""
	def __init__(
			self,
			experiences: Optional[List[Experience]] = None,
			gamma: Optional[float] = None,
	):
		self.experiences = experiences if experiences is not None else []
		self._terminal_flag = experiences[-1].terminal if experiences else False
		self.gamma = gamma

	@property
	def _default_gamma(self):
		return self.gamma if self.gamma is not None else 0.99

	def set_terminal(self, terminal: bool):
		self._terminal_flag = terminal
		if self._terminal_flag:
			self.propagate_rewards()

	def propagate_rewards(self, gamma: Optional[float] = 0.99):
		"""
		Propagate the rewards to the next experiences.
		"""
		gamma = gamma if gamma is not None else self._default_gamma
		for i in reversed(range(len(self.experiences))):
			if i == len(self.experiences) - 1:
				self.experiences[i].discounted_reward = self.experiences[i].reward
			else:
				self.experiences[i].discounted_reward = (
						self.experiences[i].reward + gamma * self.experiences[i + 1].discounted_reward
				)

	def propagate_values(self, lmbda: Optional[float] = 0.95):
		raise NotImplementedError("Not implemented yet.")

	def __iter__(self):
		return iter(self.experiences)

	def __len__(self):
		return len(self.experiences)

	def __getitem__(self, index: int) -> Experience:
		return self.experiences[index]

	def append(self, experience: Experience):
		if self._terminal_flag:
			raise ValueError("Cannot append experience to a terminal trajectory.")
		self.experiences.append(experience)
		self.set_terminal(experience.terminal)

	def append_and_terminate(self, experience: Experience):
		self.append(experience)
		self.set_terminal(True)


class ReplayBuffer:
	def __init__(self, capacity=np.inf, seed=None, use_priority=True):
		self.__capacity = capacity
		self.random_generator = np.random.RandomState(seed)
		self.data: List[Experience] = []
		self._counter = 0
		self._counter_is_started = False
		self.use_priority = use_priority

	@property
	def counter(self):
		return self._counter

	@property
	def capacity(self):
		return self.__capacity
	
	def set_seed(self, seed: int):
		self.random_generator.seed(seed)

	def start_counter(self):
		self._counter_is_started = True
		self._counter = 0

	def stop_counter(self):
		self._counter_is_started = False
		self._counter = 0

	def reset_counter(self):
		self.stop_counter()

	def increment_counter(self, increment: int = 1):
		self._counter += increment

	def increase_capacity(self, increment: int):
		self.__capacity += increment
	
	def extend(self, iterable: Iterable[Experience]):
		_ = [self.store(e) for e in iterable]
	
	def __len__(self):
		return len(self.data)

	def __iter__(self):
		return iter(self.data)

	def __getitem__(self, idx: int) -> Experience:
		return self.data[idx]
	
	def store(self, element: Experience):
		"""
		Stores an element. If the replay buffer is already full, deletes the oldest
		element to make space.
		"""
		if len(self.data) >= self.__capacity:
			if self.use_priority:
				self.data.pop(np.argmin([np.abs(e.advantage) for e in self.data]))
			else:
				self.data.pop(0)
		self.data.append(element)
		if self._counter_is_started:
			self._counter += 1
	
	def get_random_batch(self, batch_size: int) -> List[Experience]:
		"""
		Returns a list of batch_size elements from the buffer.
		"""
		return self.random_generator.choice(self.data, size=batch_size)
	
	def get_batch_tensor(self, batch_size: int, device='cpu') -> BatchExperience:
		"""
		Returns a list of batch_size elements from the buffer.
		"""
		batch = self.get_random_batch(batch_size)
		return BatchExperience(batch, device=device)
	
	def get_batch_generator(
			self,
			batch_size: int,
			n_batches: int = None,
			randomize: bool = True,
			device='cpu',
	) -> Iterable[BatchExperience]:
		"""
		Returns a generator of batch_size elements from the buffer.
		"""
		max_idx = int(batch_size * int(len(self) / batch_size))
		indexes = np.arange(max_idx).reshape(-1, batch_size)
		if n_batches is None:
			n_batches = indexes.shape[0]
		else:
			n_batches = min(n_batches, indexes.shape[0])
		if randomize:
			self.random_generator.shuffle(indexes)
		for i in range(n_batches):
			batch = [self.data[j] for j in indexes[i]]
			yield BatchExperience(batch, device=device)

	def save(self, filename: str):
		buffer_copy = deepcopy(self)
		for i, e in enumerate(buffer_copy.data):
			buffer_copy.data[i] = Experience(
				obs=e.obs,
				action=e.action.to_numpy(),
				reward=e.reward,
				terminal=e.terminal,
				next_obs=e.next_obs,
			)
		with open(filename, 'wb') as file:
			pickle.dump(self, file)

	@staticmethod
	def load(filename: str) -> 'ReplayBuffer':
		with open(filename, 'rb') as file:
			buffer = pickle.load(file)
		for i, e in enumerate(buffer.data):
			buffer.data[i] = Experience(
				obs=e.obs,
				action=TensorActionTuple.from_numpy(e.action),
				reward=e.reward,
				terminal=e.terminal,
				next_obs=e.next_obs,
			)
		return buffer



import os
import shutil
import time
import warnings
from collections import defaultdict, deque
from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from mlagents_envs.base_env import BaseEnv, DecisionSteps, TerminalSteps
from torch import nn
from tqdm.auto import tqdm

from PythonAcademy.src.base_agent import BaseAgent
from PythonAcademy.src.buffers import BatchExperience, Experience, ReplayBuffer, Trajectory
from PythonAcademy.src.checkpoints_manager import CheckpointManager, LoadCheckpointMode
from PythonAcademy.src.curriculum import Curriculum
from PythonAcademy.src.utils import TensorActionTuple, TrainingHistoriesMap, TrainingHistory, linear_decay
from PythonAcademy.src.utils import unbatch_actions


class AgentsHistoryMaps:
	"""
	Class to store the mapping between agents and their history maps

	Attributes:
		trajectories (Dict[int, Trajectory]): Mapping between agent ids and their trajectories
		last_obs (Dict[int, np.ndarray]): Mapping between agent ids and their last observations
		last_action (Dict[int, np.ndarray]): Mapping between agent ids and their last actions
		cumulative_reward (Dict[int, float]): Mapping between agent ids and their cumulative rewards
	"""

	def __init__(self, buffer: Optional[ReplayBuffer] = None):
		self.buffer = buffer if buffer is not None else ReplayBuffer()
		self.trajectories: Dict[int, Trajectory] = defaultdict(Trajectory)
		self.last_obs: Dict[int, Any] = defaultdict()
		self.last_action: Dict[int, TensorActionTuple] = defaultdict()
		self.cumulative_reward: Dict[int, float] = defaultdict(lambda: 0.0)
		self._terminal_counter = 0

	@property
	def terminals_count(self) -> int:
		"""
		:return: The number of terminal steps
		"""
		return self._terminal_counter

	def update_terminals_(
			self,
			terminal_steps: TerminalSteps
	) -> List[float]:
		"""
		Execute terminal steps and return the rewards
		:param terminal_steps: The terminal steps
		:return: The rewards
		"""
		cumulative_rewards = []
		# For all Agents with a Terminal Step:
		for agent_id_terminated in terminal_steps:
			# Create its last experience (is last because the Agent terminated)
			last_experience = Experience(
				obs=deepcopy(self.last_obs[agent_id_terminated]),
				reward=terminal_steps[agent_id_terminated].reward,
				terminal=not terminal_steps[agent_id_terminated].interrupted,
				action=self.last_action[agent_id_terminated].copy(),
				next_obs=terminal_steps[agent_id_terminated].obs,
			)
			self.trajectories[agent_id_terminated].append_and_terminate(last_experience)
			# Clear its last observation and action (Since the trajectory is over)
			self.last_obs.pop(agent_id_terminated)
			self.last_action.pop(agent_id_terminated)
			# Report the cumulative reward
			cumulative_rewards.append(
				self.cumulative_reward.pop(agent_id_terminated, 0.0)
				+ terminal_steps[agent_id_terminated].reward
			)
			# Add the Trajectory to the buffer
			self.buffer.extend(self.trajectories.pop(agent_id_terminated))
			self._terminal_counter += 1
		return cumulative_rewards

	def update_decisions_(self, decision_steps: DecisionSteps):
		"""
		Execute the decision steps of the agents
		:param decision_steps: The decision steps
		:return: None
		"""
		# For all Agents with a Decision Step:
		for agent_id_decisions in decision_steps:
			# If the Agent requesting a decision has a "last observation"
			if agent_id_decisions in self.last_obs:
				# Create an Experience from the last observation and the Decision Step
				exp = Experience(
					obs=deepcopy(self.last_obs[agent_id_decisions]),
					reward=decision_steps[agent_id_decisions].reward,
					terminal=False,
					action=self.last_action[agent_id_decisions].copy(),
					next_obs=decision_steps[agent_id_decisions].obs,
				)
				# Update the Trajectory of the Agent and its cumulative reward
				self.trajectories[agent_id_decisions].append(exp)
				self.cumulative_reward[agent_id_decisions] += decision_steps[agent_id_decisions].reward
			# Store the observation as the new "last observation"
			self.last_obs[agent_id_decisions] = decision_steps[agent_id_decisions].obs

	def update_actions_(self, actions: TensorActionTuple, decision_steps: DecisionSteps):
		"""
		Add the actions to the last observation of the agents.
		:param actions: The actions
		:param decision_steps: The decision steps
		:return: None
		"""
		actions_list = unbatch_actions(actions)
		for agent_index, agent_id in enumerate(decision_steps.agent_id):
			self.last_action[agent_id] = actions_list[agent_index]

	def update_(
			self,
			decision_steps: Optional[DecisionSteps] = None,
			terminal_steps: Optional[TerminalSteps] = None,
			actions: Optional[TensorActionTuple] = None,
	) -> Optional[List[float]]:
		"""
		Update the replay buffer with the given steps.
		:param buffer: The replay buffer to store the experiences
		:param decision_steps: The decision steps
		:param terminal_steps: The terminal steps
		:param actions: The actions
		:return: The cumulative rewards if the terminal steps are given
		"""
		cumulative_rewards = None
		if terminal_steps is not None:
			cumulative_rewards = self.update_terminals_(terminal_steps)
		if decision_steps is not None:
			self.update_decisions_(decision_steps)
		if actions is not None:
			assert decision_steps is not None, "If actions are given, decision_steps must be given"
			self.update_actions_(actions, decision_steps)
		return cumulative_rewards


class RLAcademy:
	def __init__(
			self,
			env: BaseEnv,
			agent: BaseAgent,
			behavior_name: Optional[str] = None,
			checkpoint_folder: Optional[str] = None,
			curriculum: Optional[Curriculum] = None,
			verbose: bool = True,
			**kwargs
	):
		"""
		Initialize the RL Academy.
		:param env: The environment to train the agent in.
		:param agent:
		:param behavior_name:
		:param checkpoint_folder:
		:param curriculum:
		:param kwargs:
		"""
		self.env = env
		self.behavior_name = behavior_name if behavior_name is not None else list(env.behavior_specs)[0]
		self.behavior_short_name = self.behavior_name.split('?')[0]
		self.policy = agent
		self._last_policy = self._copy_policy()
		if checkpoint_folder is None:
			checkpoint_folder = f"checkpoints_{self.policy.name}_{self.behavior_short_name}"
		self.checkpoint_folder = checkpoint_folder
		self.verbose = verbose
		self.kwargs = self._set_default_academy_kwargs(kwargs)
		self.curriculum = curriculum
		self.training_histories = TrainingHistoriesMap(self.curriculum)
		self.checkpoint_manager = CheckpointManager(
			checkpoint_folder,
			meta_path_prefix=f"{self.policy.name}_{self.behavior_short_name}"
		)

		self.policy_optimizer = torch.optim.Adam(
			self.policy.parameters(),
			lr=self.kwargs["init_lr"],
			weight_decay=self.kwargs["weight_decay"],
		)
		if self.curriculum is None:
			self.cloning_optimizer = None
		else:
			self.cloning_optimizer = torch.optim.Adam(
				self.policy.parameters(),
				lr=self.kwargs["init_lr"] * self.kwargs["bc_strength"],
				weight_decay=self.kwargs["weight_decay"],
			)
		self.continuous_criterion = nn.MSELoss()
		self.discrete_criterion = nn.MSELoss()

	@staticmethod
	def _set_default_academy_kwargs(kwargs: Dict[str, Any]) -> Dict[str, Any]:
		"""
		Set default values for the kwargs of the fit method.
		:param kwargs:
			close_env: Whether to close the environment after the training.
			n_epochs: Number of epochs to train each iteration.
			init_lr: Initial learning rate.
			min_lr: Minimum learning rate.
			weight_decay: Weight decay.
			init_epsilon: Initial epsilon. Epsilon is the probability of choosing a random action.
			epsilon_decay: Epsilon decay.
			min_epsilon: Minimum epsilon.
			gamma: Discount factor.
			tau: Target network update rate.
			n_batches: Number of batches to train each iteration.
			batch_size: Batch size.
			update_freq: Number of steps between each update.
			curriculum_strength: Strength of the teacher learning strategy.
		:return:
		"""
		kwargs.setdefault("close_env", False)
		kwargs.setdefault("n_epochs", 3)
		kwargs.setdefault("init_lr", 3.0e-4)
		kwargs.setdefault("min_lr", 3.0e-4)
		kwargs.setdefault("lr", kwargs["init_lr"])
		kwargs.setdefault("weight_decay", 1e-5)
		kwargs.setdefault("init_epsilon", 0.01)
		kwargs.setdefault("epsilon_decay", 0.995)
		kwargs.setdefault("min_epsilon", 0.0)
		kwargs.setdefault("gamma", 0.99)
		kwargs.setdefault("n_batches", 3)
		kwargs.setdefault("tau", 1/kwargs["n_batches"])
		kwargs.setdefault("batch_size", 256)
		kwargs.setdefault("update_freq", 32)
		kwargs.setdefault("bc_strength", 0.5)
		kwargs.setdefault("buffer_size", 4096)
		kwargs.setdefault("clip_ratio", 0.2)
		kwargs.setdefault("use_priority_buffer", True)

		assert kwargs["batch_size"] <= kwargs["buffer_size"]
		assert kwargs["batch_size"] > 0
		return kwargs

	def _copy_policy(self, requires_grad: bool = False) -> BaseAgent:
		"""
		Copy the policy to a new instance.
		:return: The copied policy.
		"""
		policy_copy = deepcopy(self.policy)
		for param in policy_copy.parameters():
			param.requires_grad = requires_grad
		policy_copy.eval()
		return policy_copy

	def _update_optimizer_(self):
		self.policy_optimizer = torch.optim.Adam(
			self.policy.parameters(),
			lr=self.kwargs["init_lr"],
			weight_decay=self.kwargs["weight_decay"]
		)

	def generate_trajectories(
			self,
			n_trajectories: int,
			buffer: Optional[ReplayBuffer] = None,
			epsilon: float = 0.0,
			p_bar_position: int = 0,
			verbose: Optional[bool] = None
	) -> Tuple[ReplayBuffer, List[float]]:
		if buffer is None:
			buffer = ReplayBuffer(self.kwargs["buffer_size"], use_priority=self.kwargs["use_priority_buffer"])
		if verbose is None:
			verbose = self.verbose
		self.env.reset()
		agents_history_maps = AgentsHistoryMaps(buffer)
		cumulative_rewards: List[float] = []
		p_bar = tqdm(
			total=n_trajectories, disable=not verbose, desc="Generating Trajectories", position=p_bar_position
		)
		while agents_history_maps.terminals_count < n_trajectories:  # While not enough data in the buffer
			decision_steps, terminal_steps = self.env.get_steps(self.behavior_name)
			actions = None
			if len(decision_steps) > 0:
				if np.random.random() < epsilon:
					actions = self.policy.get_random_actions(len(decision_steps))
				else:
					actions = self.policy.get_actions(decision_steps.obs)
				self.env.set_actions(self.behavior_name, actions.to_numpy())
			new_cumulative_rewards = agents_history_maps.update_(decision_steps, terminal_steps, actions)
			p_bar.update(min(len(new_cumulative_rewards), n_trajectories - len(cumulative_rewards)))
			cumulative_rewards.extend(new_cumulative_rewards)
			p_bar.set_postfix(cumulative_reward=f"{np.mean(cumulative_rewards) if cumulative_rewards else 0.0:.3f}")
			self.env._update_k_p_on_datum()
		p_bar.close()
		return buffer, cumulative_rewards

	def load_checkpoint(
			self,
			load_checkpoint_mode: LoadCheckpointMode = LoadCheckpointMode.BEST_ITR
	) -> dict:
		checkpoint = self.checkpoint_manager.load_checkpoint(load_checkpoint_mode)
		self.policy.load_state_dict(checkpoint[CheckpointManager.CHECKPOINT_STATE_DICT_KEY], strict=True)
		return checkpoint

	def plot_training_history(self, training_history: TrainingHistory = None, show: bool = False) -> str:
		if training_history is None:
			training_history = self.training_histories
		save_path = f"./{self.checkpoint_folder}/training_history.png"
		os.makedirs(f"./{self.checkpoint_folder}/", exist_ok=True)
		training_history.plot(save_path=save_path, show=show)
		return save_path

	def check_and_load_state_from_academy_checkpoint(
			self,
			load_checkpoint_mode: LoadCheckpointMode = None,
			force_overwrite: bool = False,
	) -> int:
		start_itr = 0
		if load_checkpoint_mode is None:
			if os.path.exists(self.checkpoint_manager.checkpoints_meta_path):
				if force_overwrite:
					shutil.rmtree(self.checkpoint_folder)
				else:
					raise ValueError(
						f"{self.checkpoint_manager.checkpoints_meta_path} already exists. "
						f"Set force_overwrite flag to True to overwrite existing saves."
					)
		else:
			try:
				checkpoint = self.load_checkpoint(load_checkpoint_mode)
				self.policy.load_state_dict(checkpoint[CheckpointManager.CHECKPOINT_STATE_DICT_KEY], strict=True)
				self.policy_optimizer.load_state_dict(checkpoint[CheckpointManager.CHECKPOINT_OPTIMIZER_STATE_DICT_KEY])
				start_itr = int(checkpoint[CheckpointManager.CHECKPOINT_ITR_KEY]) + 1
				self.training_histories: TrainingHistoriesMap = checkpoint[CheckpointManager.CHECKPOINT_TRAINING_HISTORY_KEY]
				temp_curriculum = self.curriculum
				self.curriculum = self.training_histories.curriculum
				if temp_curriculum is not None:
					self.curriculum.update_teachers_and_channels(temp_curriculum)
				self.plot_training_history(show=False)
			except FileNotFoundError as e:
				if self.verbose:
					warnings.warn(f"Error: {e}", Warning)
					warnings.warn("No such checkpoint. Fit from beginning.")
		return start_itr

	def _init_train_buffer(self) -> ReplayBuffer:
		self.env.reset()
		if self.curriculum is not None:
			self.curriculum.on_iteration_start()
		buffer, _ = self.generate_trajectories(self.kwargs["batch_size"])
		return buffer

	def train(
			self,
			n_iterations: int = int(1e6),
			load_checkpoint_mode: LoadCheckpointMode = None,
			force_overwrite: bool = False,
			save_freq: int = int(1e2),
			max_seconds: float = np.inf,
			**kwargs
	) -> Union[TrainingHistory, TrainingHistoriesMap]:
		"""
		Train the agent in the given environment.
		:param n_iterations: The number of iterations to train the agent.
		:param load_checkpoint_mode: The mode to use when loading a checkpoint.
		:param force_overwrite: True to overwrite the checkpoint if it already exists.
		:param save_freq: The frequency with which to save the checkpoint.
		:param max_seconds: The maximum number of seconds to train for.
		:param kwargs: The fit kwargs.
		:return: The training history.
		"""
		# TODO: add learning rate decay
		self.kwargs.update(kwargs)
		self._update_optimizer_()

		start_time = time.time()
		start_itr = self.check_and_load_state_from_academy_checkpoint(load_checkpoint_mode, force_overwrite)
		best_rewards = self.training_histories.max("Rewards")
		best_saved_rewards = best_rewards
		buffer = self._init_train_buffer()
		self.env.reset()
		last_save_rewards = deque(maxlen=save_freq)
		p_bar = tqdm(range(start_itr, n_iterations), disable=not self.verbose, desc="Training")
		for i in p_bar:
			if self.curriculum is not None:
				self.curriculum.on_iteration_start()
			break_flag = False
			epsilon = linear_decay(
				self.kwargs["init_epsilon"], self.kwargs["min_epsilon"], self.kwargs["epsilon_decay"], i
			)
			teacher_loss = self.fit_curriculum_buffer()
			itr_metrics = self._exec_fit_itr_(epsilon, buffer)
			best_rewards = max(best_rewards, itr_metrics["Rewards"])
			last_save_rewards.append(itr_metrics["Rewards"])
			p_bar_postfix = {
				"loss": f"{itr_metrics['Loss']:.3f}",
				"itr_rewards": f"{itr_metrics['Rewards']:.3f}",
				f"last_{save_freq}_rewards": f"{np.mean(last_save_rewards):.3f}",
				"best_rewards": f"{best_rewards:.3f}",
			}
			if teacher_loss is not None:
				itr_metrics.update(TeacherLoss=teacher_loss)
				p_bar_postfix.update(TeacherLoss=f"{teacher_loss:.3f}")
			self.training_histories.concat(itr_metrics)
			if self.curriculum is not None:
				curriculum_out = self.curriculum.on_iteration_end(itr_metrics)
				p_bar_postfix.update(curriculum_out.messages)
				break_flag = self.curriculum.is_completed
				if curriculum_out.lesson_completed:
					best_saved_rewards = -np.inf
					best_rewards = -np.inf
					self._update_bc_optimizer_()
			p_bar.set_postfix(p_bar_postfix)
			if time.time() - start_time > max_seconds:
				warnings.warn(f"Training stopped after {max_seconds} seconds.")
				break_flag = True
			if i % save_freq == 0 or break_flag or i == n_iterations - 1:
				best_saved_rewards = self._make_itr_checkpoint(i, itr_metrics, best_saved_rewards)
			if break_flag:
				break
		p_bar.close()
		if self.kwargs.get("close_env", True):
			self.env.close()
		self.plot_training_history(show=False)

		if self.curriculum is None:
			return self.training_histories.report_history
		return self.training_histories

	def _make_itr_checkpoint(self, itr: int, itr_metrics: Dict[str, float], best_saved_rewards: float) -> float:
		is_best = itr_metrics["Rewards"] > best_saved_rewards
		if is_best:
			best_saved_rewards = itr_metrics["Rewards"]
		self.checkpoint_manager.save_checkpoint(
			itr, itr_metrics, is_best,
			state_dict=self.policy.state_dict(),
			optimizer_state_dict=self.policy_optimizer.state_dict(),
			training_history=self.training_histories,
		)
		self.plot_training_history(show=False)
		return best_saved_rewards

	def _update_bc_optimizer_(self):
		if self.curriculum.teacher_buffer is None:
			return None
		base_lr = self.kwargs["lr"]
		bc_strength = self.kwargs["bc_strength"]
		if self.curriculum.current_lesson.teacher_strength is not None:
			bc_strength = self.curriculum.current_lesson.teacher_strength
		for g in self.cloning_optimizer.param_groups:
			g['lr'] = base_lr * bc_strength

	def _exec_fit_itr_(
			self,
			epsilon: float,
			buffer: ReplayBuffer,
	) -> Dict[str, float]:
		buffer, cumulative_rewards = self.generate_trajectories(
			self.kwargs["update_freq"], buffer, epsilon, p_bar_position=0, verbose=False,
		)
		cum_rewards = np.mean(cumulative_rewards)
		itr_loss = self.fit_buffer(buffer)
		return dict(Rewards=cum_rewards, Loss=itr_loss)

	def fit_curriculum_buffer(self) -> Optional[float]:
		"""
		Fit the curriculum buffer.
		:return: The teacher loss.
		"""
		buffer = self.curriculum.teacher_buffer
		if buffer is None:
			return None
		batch_size = min(len(buffer), self.kwargs["batch_size"])
		batches = buffer.get_batch_generator(
			batch_size, self.kwargs["n_batches"], randomize=True, device=self.policy.device
		)
		losses = []
		for _ in range(self.kwargs["n_epochs"]):
			for batch in batches:
				loss = self._behaviour_cloning_update_weights(batch)
				losses.append(loss)
		self._last_policy = self._copy_policy(requires_grad=False)
		return float(np.mean(losses))

	def fit_buffer(
			self,
			buffer: ReplayBuffer,
	) -> float:
		"""
		Fit the agent on the given buffer.
		:param buffer: The replay buffer.
		:return: The loss.
		"""
		batch_size = min(len(buffer), self.kwargs["batch_size"])
		batches = buffer.get_batch_generator(
			batch_size, self.kwargs["n_batches"], randomize=True, device=self.policy.device
		)
		losses = []
		for _ in range(self.kwargs["n_epochs"]):
			for batch in batches:
				# predictions = self.policy.get_actions(batch.obs, as_numpy=False)
				# targets = self._last_policy.get_actions(batch.next_obs, as_numpy=False)
				# loss = self.update_weights(batch, predictions, targets)
				# self._last_policy.soft_update(self.policy, tau=self.kwargs["tau"])

				loss = self.update_policy_weights(batch)
				# self._last_policy = self._copy_policy(requires_grad=False)
				self._last_policy.soft_update(self.policy, tau=self.kwargs["tau"])
				losses.append(loss)
		self._last_policy = self._copy_policy(requires_grad=False)
		return float(np.mean(losses))

	def _behaviour_cloning_compute_continuous_loss(self, batch: BatchExperience, predictions) -> torch.Tensor:
		targets = batch.continuous_actions
		if torch.numel(batch.continuous_actions) == 0:
			continuous_loss = 0.0
		else:
			continuous_loss = self.continuous_criterion(predictions, targets.to(self.policy.device))
		return continuous_loss

	def _behaviour_cloning_compute_discrete_loss(self, batch: BatchExperience, predictions) -> torch.Tensor:
		targets = batch.discrete_actions
		if torch.numel(batch.discrete_actions) == 0:
			discrete_loss = 0.0
		else:
			warnings.warn("Discrete loss is not implemented with cross entropy loss. This is a temporary solution.")
			discrete_loss = self.discrete_criterion(predictions, targets.to(self.policy.device))
		return discrete_loss

	def _behaviour_cloning_update_weights(
			self,
			batch: BatchExperience,
	) -> float:
		assert torch.numel(batch.continuous_actions) + torch.numel(batch.discrete_actions) > 0
		predictions = self.policy.get_actions(batch.obs)
		bc_continuous_loss = self._behaviour_cloning_compute_continuous_loss(batch, predictions.continuous)
		bc_discrete_loss = self._behaviour_cloning_compute_discrete_loss(batch, predictions.discrete)
		bc_loss = bc_continuous_loss + bc_discrete_loss
		# Perform the backpropagation
		self.cloning_optimizer.zero_grad()
		bc_loss.backward()
		self.cloning_optimizer.step()
		return bc_loss.detach().cpu().numpy().item()

	def _compute_continuous_loss(self, batch: BatchExperience, predictions, targets) -> torch.Tensor:
		if torch.numel(batch.continuous_actions) == 0:
			continuous_loss = 0.0
		else:
			targets = (
					batch.rewards
					+ (1.0 - batch.terminals)
					* self.kwargs["gamma"]
					* targets
			).to(self.policy.device)
			continuous_loss = self.continuous_criterion(predictions, targets)
		return continuous_loss

	def _compute_discrete_loss(self, batch: BatchExperience, predictions, targets) -> torch.Tensor:
		if torch.numel(batch.discrete_actions) == 0:
			discrete_loss = 0.0
		else:
			targets = (
					batch.rewards
					+ (1.0 - batch.terminals)
					* self.kwargs["gamma"]
					* targets
			).to(self.policy.device)
			warnings.warn("Discrete loss is not implemented with cross entropy loss. This is a temporary solution.")
			discrete_loss = self.discrete_criterion(predictions, targets)
		return discrete_loss

	def _compute_proximal_policy_optimization_loss(self, batch: BatchExperience, predictions, targets) -> torch.Tensor:
		if torch.numel(batch.continuous_actions) == 0:
			continuous_loss = 0.0
		else:
			targets = (
					batch.rewards
					+ (1.0 - batch.terminals)
					* self.kwargs["gamma"]
					* targets
			).to(self.policy.device)
			continuous_loss = self.continuous_criterion(predictions, targets)
		return continuous_loss

	def update_weights(
			self,
			batch: BatchExperience,
			predictions: TensorActionTuple,
			targets: TensorActionTuple,
	) -> float:
		warnings.warn("This method is deprecated. Please use update_policy_weights instead.", DeprecationWarning)
		"""
		Performs a single update of the Q-Network using the provided optimizer and buffer
		"""
		assert torch.numel(batch.continuous_actions) + torch.numel(batch.discrete_actions) > 0
		continuous_loss = self._compute_continuous_loss(batch, predictions.continuous, targets.continuous)
		discrete_loss = self._compute_discrete_loss(batch, predictions.discrete, targets.discrete)
		loss = continuous_loss + discrete_loss
		# Perform the backpropagation
		self.policy_optimizer.zero_grad()
		loss.backward()
		self.policy_optimizer.step()
		return loss.detach().cpu().numpy().item()

	def _compute_policy_ratio(self, batch: BatchExperience) -> torch.Tensor:
		policy_predictions = self.policy.get_logits(batch.obs)
		last_policy_predictions = self._last_policy.get_logits(batch.obs)
		return policy_predictions / (last_policy_predictions + 1e-8)

	def _compute_policy_loss(self, batch: BatchExperience) -> torch.Tensor:
		policy_ratio = self._compute_policy_ratio(batch)
		policy_loss = -torch.mean(
			torch.minimum(
				policy_ratio * batch.advantages,
				torch.clamp(
					policy_ratio,
					1 - self.kwargs["clip_ratio"],
					1 + self.kwargs["clip_ratio"]
				) * batch.advantages
			)
		)
		return policy_loss

	def update_policy_weights(self, batch: BatchExperience) -> float:
		"""
		Performs a single update of the policy network using the provided optimizer and buffer
		"""
		policy_loss = self._compute_policy_loss(batch)
		# Perform the backpropagation
		self.policy_optimizer.zero_grad()
		policy_loss.backward()
		self.policy_optimizer.step()
		return policy_loss.detach().cpu().numpy().item()

	def close(self):
		self.env.close()



from typing import Dict, Iterable, Optional, Union
import pythonbasictools as pybt
import torch

from ..callbacks.base_callback import BaseCallback
from ..utils import filter_parameters


class BaseRegularization(torch.nn.Module, BaseCallback):
	"""
	Base class for regularization.

	:Attributes:
		- :attr:`params` (torch.nn.ParameterList): The parameters which are regularized.
		- :attr:`Lambda` (float): The weight of the regularization. In other words, the coefficient that multiplies the loss.

	"""
	def __init__(
			self,
			params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
			Lambda: float = 1.0,
			optimizer: Optional[torch.optim.Optimizer] = None,
			**kwargs
	):
		"""
		Constructor of the BaseRegularization class.
		
		:param params: The parameters which are regularized.
		:type params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]]
		:param Lambda: The weight of the regularization. In other words, the coefficient that multiplies the loss.
		:type Lambda: float
		"""
		super().__init__(**kwargs)
		BaseCallback.__init__(self, **kwargs)
		if isinstance(params, dict):
			self.params = list(params.values())
		else:
			self.params = list(params)
		self.params = torch.nn.ParameterList(self.params)
		self.Lambda = Lambda
		self.name = self.__class__.__name__
		self.optimizer = optimizer

	def __call__(self, *args, **kwargs) -> torch.Tensor:
		"""
		Call the forward pass of the regularization and scale it by the :attr:`Lambda` attribute.
		
		:param args: args of the forward pass.
		:param kwargs: kwargs of the forward pass.
		
		:return: The loss of the regularization.
		:rtype: torch.Tensor
		"""
		out = super(BaseRegularization, self).__call__(*args, **kwargs)
		return self.Lambda * out

	def forward(self, *args, **kwargs) -> torch.Tensor:
		"""
		Compute the forward pass of the regularization.
		
		:param args: args of the forward pass.
		:param kwargs: kwargs of the forward pass.
		
		:return: The loss of the regularization.
		:rtype: torch.Tensor
		"""
		raise NotImplementedError("forward method must be implemented")
	
	def on_optimization_end(self, trainer, **kwargs):
		reg_loss = self()
		if self.optimizer is not None:
			self.optimizer.zero_grad()
		reg_loss.backward()
		if self.optimizer is not None:
			self.optimizer.step()


class RegularizationList(BaseRegularization):
	"""
	Regularization that applies a list of regularization.
	
	:Attributes:
		- :attr:`regularizations` (Iterable[BaseRegularization]): The regularizations to apply.
	"""
	def __init__(
			self,
			regularizations: Optional[Iterable[BaseRegularization]] = None,
			optimizer: Optional[torch.optim.Optimizer] = None,
			**kwargs
	):
		"""
		Constructor of the RegularizationList class.
		
		:param regularizations: The regularizations to apply.
		:type regularizations: Optional[Iterable[BaseRegularization]]
		"""
		self.regularizations = regularizations if regularizations is not None else []
		_params = []
		for regularization in self.regularizations:
			_params.extend(regularization.params)
		super(RegularizationList, self).__init__(
			params=_params,
			Lambda=1.0,
			optimizer=optimizer,
			**kwargs
		)
		self.regularizations = regularizations if regularizations is not None else []

	def __iter__(self):
		"""
		Iterate over the regularizations.
		
		:return: An iterator over the regularizations.
		"""
		return iter(self.regularizations)

	def forward(self, *args, **kwargs) -> torch.Tensor:
		"""
		Compute the forward pass of the regularization.
		
		:param args: args of the forward pass.
		:param kwargs: kwargs of the forward pass.
		"""
		if len(self.regularizations) == 0:
			return torch.tensor(0)
		loss = sum([regularization(*args, **kwargs) for regularization in self.regularizations])
		return loss


# @pybt.docstring.inherit_fields_docstring(fields=["Attributes"], bases=[BaseRegularization])
class Lp(BaseRegularization):
	"""
	Regularization that applies LP norm.
	
	:Attributes:
		- :attr:`p` (int): The p parameter of the LP norm. Example: p=1 -> L1 norm, p=2 -> L2 norm.
	"""
	def __init__(
			self,
			params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
			Lambda: float = 1.0,
			p: int = 1,
			**kwargs
	):
		"""
		Constructor of the L1 class.
		
		:param params: The parameters which are regularized.
		:type params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]]
		:param Lambda: The weight of the regularization. In other words, the coefficient that multiplies the loss.
		:type Lambda: float
		:param p: The p parameter of the LP norm. Example: p=1 -> L1 norm, p=2 -> L2 norm.
		:type p: int
		"""
		super(Lp, self).__init__(params, Lambda, **kwargs)
		self.p = p

	def forward(self, *args, **kwargs) -> torch.Tensor:
		"""
		Compute the forward pass of the regularization.
		
		:param args: args of the forward pass
		:param kwargs: kwargs of the forward pass
		
		:return: The loss of the regularization.
		:rtype: torch.Tensor
		"""
		loss = torch.tensor(0.0, requires_grad=True).to(self.params[0].device)
		for param in filter_parameters(self.params, requires_grad=True):
			loss += torch.linalg.norm(param, self.p)
		return loss


# @pybt.docstring.inherit_fields_docstring(fields=["Attributes"], bases=[Lp])
class L1(Lp):
	"""
	Regularization that applies L1 norm.
	"""
	def __init__(
			self,
			params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
			Lambda: float = 1.0,
			**kwargs
	):
		"""
		Constructor of the L1 class.
		
		:param params: The parameters which are regularized.
		:type params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]]
		:param Lambda: The weight of the regularization. In other words, the coefficient that multiplies the loss.
		:type Lambda: float
		"""
		super(L1, self).__init__(params, Lambda, p=1, **kwargs)


# @pybt.docstring.inherit_fields_docstring(fields=["Attributes"], bases=[Lp])
class L2(Lp):
	"""
	Regularization that applies L2 norm.
	"""
	def __init__(
			self,
			params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
			Lambda: float = 1.0,
			**kwargs
	):
		"""
		Constructor of the L2 class.
		
		:param params: The parameters which are regularized.
		:type params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]]
		:param Lambda: The weight of the regularization. In other words, the coefficient that multiplies the loss.
		:type Lambda: float
		"""
		super(L2, self).__init__(params, Lambda, p=2, **kwargs)





import enum
import os

import numpy as np
import torch
from torchvision.transforms import Compose, ToTensor


class ImgToSpikes:
	def __init__(
			self,
			n_steps: int,
			t_max: float = None,
			tau=20.0 * 1e-3,
			thr=0.2,
			use_periods=False,
			epsilon=1e-7,
	):
		"""
		:param n_steps: The number of time step
		:param t_max:
		:param tau: The membrane time constant of the LIF neuron to be charged
		:param thr: The firing threshold value
		:param epsilon: A generic (small) epsilon > 0
		"""
		self.n_steps = n_steps
		self.t_max = n_steps if t_max is None else t_max
		self.tau = tau
		self.thr = thr
		self.epsilon = epsilon
		self.spikes_indices = None
		self.use_periods = use_periods
		self.spikes_gen_func = self.firing_periods_to_spikes if use_periods else self.firing_times_to_spikes

	def pixels_to_firing_periods(self, x: np.ndarray) -> np.ndarray:
		"""
		Computes first firing time latency for a current input x assuming the charge time of a current based LIF neuron.
		:param x: the normalized between 0.0 and 1.0 input pixels
		:return: Time between spikes for each pixel of x
		"""
		# t_per_step = self.t_max / self.n_steps
		idx = x < self.thr
		x = np.clip(x, self.thr + self.epsilon, 1.0e9)
		T = self.tau * np.log(x / (x - self.thr))
		T[idx] = self.t_max
		# periods = T // t_per_step
		return T.astype(int)

	def firing_periods_to_spikes_loop(self, firing_periods: np.ndarray) -> np.ndarray:
		spikes = np.zeros((self.n_steps, *firing_periods.shape), dtype=float)
		starts = np.clip(firing_periods, 0, self.n_steps - 1).astype(int)
		for i, period in enumerate(firing_periods):
			spikes[np.arange(starts[i], self.n_steps, step=period, dtype=int), i] = 1.0
		return spikes

	def firing_periods_to_spikes_clip(self, firing_periods: np.ndarray) -> np.ndarray:
		if self.spikes_indices is None:
			self.spikes_indices = np.indices((self.n_steps, *firing_periods.shape))
		starts = np.clip(firing_periods, 0, self.n_steps - 1, dtype=int)
		# starts[starts > (self.n_steps - 1)] = self.n_steps - 1
		spikes_range = self.spikes_indices[0] - starts[self.spikes_indices[1]]
		spikes = ((spikes_range % firing_periods[self.spikes_indices[1]]) == 0) * (spikes_range >= 0)
		return spikes.astype(float)

	def firing_periods_to_spikes(self, firing_periods: np.ndarray) -> np.ndarray:
		if self.spikes_indices is None:
			self.spikes_indices = np.indices((self.n_steps, *firing_periods.shape))
		firing_periods[firing_periods > (self.n_steps - 1)] = self.n_steps - 1
		firing_periods[firing_periods < 1] = 1
		spikes_range = self.spikes_indices[0] - firing_periods[self.spikes_indices[1]]
		spikes = ((spikes_range % firing_periods[self.spikes_indices[1]]) == 0) * (spikes_range >= 0)
		return spikes.astype(float)

	def firing_times_to_spikes(self, firing_times: np.ndarray) -> np.ndarray:
		spikes = np.zeros((self.n_steps, *firing_times.shape))
		firing_times_mask = firing_times < self.n_steps
		pix_indexes_masked = np.arange(len(firing_times))[firing_times_mask]
		spikes[firing_times[firing_times_mask], pix_indexes_masked] = 1.
		return spikes

	def _format_inputs(self, x) -> np.ndarray:
		if isinstance(x, torch.Tensor):
			return x.numpy()
		return x

	def __call__(self, x) -> torch.Tensor:
		x = self._format_inputs(x)
		firing_periods: np.ndarray = self.pixels_to_firing_periods(x)
		spikes = self.spikes_gen_func(firing_periods)
		return torch.tensor(spikes)




from setuptools import setup
from src.neurotorch import __author__, __url__, __email__, __version__, __license__
import setuptools
from src import neurotorch

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as fh:
    install_requires = fh.readlines()

setup(
    name='NeuroTorch',
    version=__version__,
    description=neurotorch.__doc__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=__url__,
    author=__author__,
    author_email=__email__,
    license=__license__,
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.9",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=install_requires,
)


# build library
#  setup.py sdist bdist_wheel

# publish on PyPI
#   twine check dist/*
#   twine upload --repository-url https://test.pypi.org/legacy/ dist/*
#   twine upload dist/*




from typing import Callable, Dict, List, Optional, Union, Any

import torch

from . import Trainer
from ..metrics import RegressionMetrics


class RegressionTrainer(Trainer):
	def __init__(self, *args, **kwargs):
		kwargs.setdefault("predict_method", "get_prediction_trace")
		super().__init__(*args, **kwargs)
	
	@staticmethod
	def _set_default_kwargs(kwargs: Dict[str, Any]) -> Dict[str, Any]:
		kwargs = Trainer._set_default_kwargs(kwargs)
		kwargs.setdefault("foresight_time_steps", None)
		return kwargs
	
	def _set_default_criterion(self, criterion: Optional[torch.nn.Module]) -> torch.nn.Module:
		if criterion is None:
			if isinstance(self.model.output_sizes, dict):
				criterion = {
					k: torch.nn.MSELoss() for k in self.model.output_sizes
				}
			elif isinstance(self.model.output_sizes, int):
				criterion = torch.nn.MSELoss()
			else:
				raise ValueError("Unknown criterion type")
		return criterion

	def _set_default_metrics(self, metrics: Optional[List[Callable]]):
		if metrics is None:
			metrics = [RegressionMetrics(self.model)]
		return metrics





from typing import Callable

import torch


class CallableToModuleWrapper(torch.nn.Module):
	def __init__(self, callable_object: Callable):
		super().__init__()
		assert callable(callable_object), "The callable object must be callable."
		self.callable_object = callable_object
	
	def __repr__(self):
		return f"CallableToModuleWrapper({self.callable_object})"

	def forward(self, *args, **kwargs):
		return self.callable_object(*args, **kwargs)





import collections.abc
import hashlib
import pickle
import time
import warnings
from typing import Callable, Dict, List, Any, Tuple, Union, Iterable, Optional, Sequence

import numpy as np
import torch
import torchvision
from matplotlib import pyplot as plt
from unstable import unstable

from .transforms.base import to_tensor


def batchwise_temporal_decay(x: torch.Tensor, decay: float = 0.9):
	r"""
	
	Apply a decay filter to the input tensor along the temporal dimension.
	
	:param x: Input of shape (batch_size, time_steps, ...).
	:type x: torch.Tensor
	:param decay: Decay factor of the filter.
	:type decay: float
	
	:return: Filtered input of shape (batch_size, ...).
	"""
	batch_size, time_steps, *_ = x.shape
	assert time_steps >= 1
	
	powers = torch.arange(time_steps, dtype=torch.float32, device=x.device).flip(0)
	weighs = torch.pow(decay, powers)
	
	x = torch.mul(x, weighs.unsqueeze(0).unsqueeze(-1))
	x = torch.sum(x, dim=1)
	return x


@unstable
def batchwise_temporal_filter(x: torch.Tensor, decay: float = 0.9):
	r"""

	Apply a low-pass filter to the input tensor along the temporal dimension.

	.. math::
		\begin{equation}\label{eqn:low-pass-filter}
			\mathcal{F}_\alpha\qty(x^t) = \alpha\mathcal{F}_\alpha\qty(x^{t-1}) + x^t.
		\end{equation}
		:label: eqn:low-pass-filter

	:param x: Input of shape (batch_size, time_steps, ...).
	:type x: torch.Tensor
	:param decay: Decay factor of the filter.
	:type decay: float

	:return: Filtered input of shape (batch_size, time_steps, ...).
	"""
	warnings.warn(
		"This function is supposed to compute the same result as `batchwise_temporal_recursive_filter` but it "
		"doesn't. Use `batchwise_temporal_recursive_filter` instead.",
		DeprecationWarning
	)
	batch_size, time_steps, *_ = x.shape
	assert time_steps >= 1
	
	# TODO: check if this is correct
	powers = torch.arange(time_steps, dtype=torch.float32, device=x.device).flip(0)
	weighs = torch.pow(decay, powers)
	
	y = torch.mul(x, weighs.unsqueeze(0).unsqueeze(-1))
	y = torch.cumsum(y, dim=1)
	return y


def batchwise_temporal_recursive_filter(x, decay: float = 0.9):
	r"""
	Apply a low-pass filter to the input tensor along the temporal dimension recursively.

	.. math::
		\begin{equation}\label{eqn:low-pass-filter}
			\mathcal{F}_\alpha\qty(x^t) = \alpha\mathcal{F}_\alpha\qty(x^{t-1}) + x^t.
		\end{equation}
		:label: eqn:low-pass-filter

	:param x: Input of shape (batch_size, time_steps, ...).
	:type x: torch.Tensor
	:param decay: Decay factor of the filter.
	:type decay: float

	:return: Filtered input of shape (batch_size, time_steps, ...).
	"""
	y = to_tensor(x).detach().clone()
	batch_size, time_steps, *_ = x.shape
	assert time_steps >= 1
	fx = 0.0
	for t in range(time_steps):
		fx = decay * fx + y[:, t]
		y[:, t] = fx
	return y


def mapping_update_recursively(d, u):
	"""
	from https://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth
	
	:param d: mapping item that wil be updated
	:param u: mapping item updater
	
	:return: updated mapping recursively
	"""
	for k, v in u.items():
		if isinstance(v, collections.abc.Mapping):
			d[k] = mapping_update_recursively(d.get(k, {}), v)
		else:
			d[k] = v
	return d


def plot_confusion_matrix(cm, classes, ):
	import matplotlib.pyplot as plt
	import itertools
	
	plt.imshow(cm, interpolation='nearest', cmap=plt.cm.Blues)
	plt.title('Confusion matrix')
	plt.colorbar()
	tick_marks = np.arange(len(classes))
	plt.xticks(tick_marks, classes, rotation=45)
	plt.yticks(tick_marks, classes)
	
	fmt = 'd'
	thresh = cm.max() / 2.
	for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
		plt.text(
			j, i,
			format(cm[i, j], fmt),
			horizontalalignment="center",
			color="white" if cm[i, j] > thresh else "black"
		)
	
	plt.ylabel('True label')
	plt.xlabel('Predicted label')
	plt.tight_layout()
	plt.show()


def legend_without_duplicate_labels_(ax: plt.Axes):
	handles, labels = ax.get_legend_handles_labels()
	unique = [(h, l) for i, (h, l) in enumerate(zip(handles, labels)) if l not in labels[:i]]
	ax.legend(*zip(*unique))


def linear_decay(init_value, min_value, decay_value, current_itr):
	return max(init_value * decay_value**current_itr, min_value)


def get_meta_name(params: Dict[str, Any]):
	meta_name = f""
	keys = sorted(list(params.keys()))
	for k in keys:
		meta_name += f"{k}-{params[k]}_"
	return meta_name[:-1]


def hash_params(params: Dict[str, Any]):
	"""
	Hash the parameters to get a unique and persistent id.
	
	:param params: The parameters to hash.
	
	:return: The hash of the parameters.
	"""
	return int(hashlib.md5(get_meta_name(params).encode('utf-8')).hexdigest(), 16)


def ravel_compose_transforms(
		transform: Union[List, Tuple, torchvision.transforms.Compose, Callable, torch.nn.ModuleList]
) -> List[Callable]:
	transforms = []
	if isinstance(transform, torchvision.transforms.Compose):
		for t in transform.transforms:
			transforms.extend(ravel_compose_transforms(t))
	elif isinstance(transform, (List, Tuple)):
		for t in transform:
			transforms.extend(ravel_compose_transforms(t))
	elif isinstance(transform, torch.nn.Module) and isinstance(transform, Iterable):
		for t in transform:
			transforms.extend(ravel_compose_transforms(t))
	elif callable(transform):
		transforms.append(transform)
	else:
		raise ValueError(f"Unsupported transform type: {type(transform)}")
	return transforms


def save_params(params: Dict[str, Any], save_path: str):
	"""
	Save the parameters in a file.
	
	:param save_path: The path to save the parameters.
	:param params: The parameters to save.
	
	:return: The path to the saved parameters.
	"""
	pickle.dump(params, open(save_path, "wb"))
	return save_path


def get_all_params_combinations(params_space: Dict[str, Any]) -> List[Dict[str, Any]]:
	"""
	Get all possible combinations of parameters.
	
	:param params_space: Dictionary of parameters.
	
	:return: List of dictionaries of parameters.
	"""
	import itertools
	# get all the combinaison of the parameters
	all_params = list(params_space.keys())
	all_params_values = list(params_space.values())
	all_params_combinaison = list(map(lambda x: list(x), list(itertools.product(*all_params_values))))
	
	# create a list of dict of all the combinaison
	all_params_combinaison_dict = list(map(lambda x: dict(zip(all_params, x)), all_params_combinaison))
	return all_params_combinaison_dict


def set_seed(seed: int):
	"""
	Set the seed of the random number generator.
	
	:param seed: The seed to set.
	"""
	import random
	import torch
	random.seed(seed)
	np.random.seed(seed)
	torch.manual_seed(seed)


def list_of_callable_to_sequential(callable_list: List[Callable]) -> torch.nn.Sequential:
	"""
	Convert a list of callable to a list of modules.
	
	:param callable_list: List of callable.
	
	:return: List of modules.
	"""
	from neurotorch.transforms.wrappers import CallableToModuleWrapper
	return torch.nn.Sequential(*[
			c if isinstance(c, torch.nn.Module) else CallableToModuleWrapper(c)
			for c in callable_list
		])


def format_pseudo_rn_seed(seed: Optional[int] = None) -> int:
	"""
	Format the pseudo random number generator seed. If the seed is None, return a pseudo random seed
	else return the given seed.
	
	:param seed: The seed to format.
	:type seed: int or None
	
	:return: The formatted seed.
	:rtype: int
	"""
	import random
	if seed is None:
		seed = int(time.time()) + random.randint(0, np.iinfo(int).max)
	assert isinstance(seed, int), "Seed must be an integer."
	return seed


def sequence_get(__sequence: Sequence, idx: int, default: Any = None) -> Any:
	try:
		return __sequence[idx]
	except IndexError:
		return default


def list_insert_replace_at(__list: List, idx: int, value: Any):
	"""
	Insert a value at a specific index. If there is already a value at this index, replace it.
	
	:param __list: The list to modify.
	:param idx: The index to insert the value.
	:param value: The value to insert.
	"""
	if idx < len(__list):
		__list[idx] = value
	else:
		__list.extend([None] * (idx - len(__list)))
		__list.append(value)


def zero_grad_params(params: Iterable[torch.nn.Parameter]):
	"""
	Set the gradient of the parameters to zero.
	
	:param params: The parameters to set the gradient to zero.
	"""
	for p in params:
		if p.grad is not None:
			p.grad.detach_()
			p.grad.zero_()


def compute_jacobian(
		*,
		model: Optional[torch.nn.Module] = None,
		params: Optional[Iterable[torch.nn.Parameter]] = None,
		x: Optional[torch.Tensor] = None,
		y: Optional[torch.Tensor] = None,
		strategy: str = "slow",
):
	"""
	Compute the jacobian of the model with respect to the parameters.
	
	# TODO: check https://medium.com/@monadsblog/pytorch-backward-function-e5e2b7e60140
	# TODO: see https://pytorch.org/tutorials/beginner/blitz/autograd_tutorial.html#sphx-glr-beginner-blitz-autograd-tutorial-py
	
	:param model: The model to compute the jacobian.
	:param params: The parameters to compute the jacobian with respect to. If None, compute the jacobian
		with respect to all the parameters of the model.
	:param x: The input to compute the jacobian. If None, use y instead.
	:param y: The output to compute the jacobian. If None, use x instead.
	:param strategy: The strategy to use to compute the jacobian. Can be "slow" or "fast". At this time the only
		strategy implemented is "slow".
	
	:return: The jacobian.
	"""
	if params is None:
		assert model is not None, "If params is None, model must be provided."
		params = model.parameters()
	zero_grad_params(params)
	
	if y is not None:
		if strategy.lower() == "fast":
			y.backward(torch.ones_like(y))
			jacobian = [p.grad.view(-1) for p in params]
		elif strategy.lower() == "slow":
			jacobian = [[] for _ in range(len(list(params)))]
			grad_outputs = torch.eye(y.shape[-1], device=y.device)
			for i in range(y.shape[-1]):
				zero_grad_params(params)
				y.backward(grad_outputs[i], retain_graph=True)
				for p_idx, param in enumerate(params):
					jacobian[p_idx].append(param.grad.view(-1).detach().clone())
			jacobian = [torch.stack(jacobian[i], dim=-1).T for i in range(len(list(params)))]
		else:
			raise ValueError(f"Unsupported strategy: {strategy}")
	elif x is not None:
		jacobian = torch.autograd.functional.jacobian(model, x, params)
	else:
		raise ValueError("Either x or y must be provided.")
	return jacobian


def vmap(f):
	# TODO: replace by torch.vmap when it is available
	def wrapper(batch_tensor):
		return torch.stack([f(batch_tensor[i]) for i in range(batch_tensor.shape[0])])
	return wrapper


def unpack_out_hh(out):
	"""
	Unpack the output of a recurrent network.
	
	:param out: The output of a recurrent network.
	
	:return: The output of the recurrent network with the hidden state.
				If there is no hidden state, consider it as None.
	"""
	out_tensor, hh = None, None
	if isinstance(out, (tuple, list)):
		if len(out) == 2:
			out_tensor, hh = out
		elif len(out) == 1:
			out_tensor = out[0]
		elif len(out) > 2:
			out_tensor, *hh = out
	else:
		out_tensor = out
	
	return out_tensor, hh


def filter_parameters(
		parameters: Union[Sequence[torch.nn.Parameter], torch.nn.ParameterList],
		requires_grad: bool = True
) -> List[torch.nn.Parameter]:
	"""
	Filter the parameters by their requires_grad attribute.
	
	:param parameters: The parameters to filter.
	:param requires_grad: The value of the requires_grad attribute to filter.
	
	:return: The filtered parameters.
	"""
	return [p for p in parameters if p.requires_grad == requires_grad]


import enum
from typing import Iterable, List, Optional, Union


class DimensionProperty(enum.Enum):
	"""
	Enum for dimension properties.

	NONE: No dimension property. This type of dimension can be used for features, neurons, unknown, etc.
	TIME: Time dimension. This type of dimension can be used for time series.
	SPATIAL: Spatial dimension. This type of dimension can be used for spatial data like images, videos, etc.
	"""
	NONE = 0
	TIME = 1
	SPATIAL = 2


class Dimension:
	"""
	This object is used to represent a dimension.

	Attributes:
		size (int): The size of the dimension.
		dtype (DimensionProperty): The type of the dimension.
		name (str): The name of the dimension.

	"""

	def __init__(
			self,
			size: Optional[int] = None,
			dtype: DimensionProperty = DimensionProperty.NONE,
			name: str = None,
	):
		"""
		Constructor for Dimension.

		:param size: The size of the dimension.
		:type size: int
		:param dtype: The type of the dimension.
		:type dtype: DimensionProperty
		:param name: The name of the dimension.
		:type name: str
		"""
		self.size: Optional[int] = size
		self.dtype: DimensionProperty = dtype
		self.name: str = name if name is not None else dtype.name

	def __eq__(self, other: Union[int, 'Dimension']) -> bool:
		"""
		Check if the dimension is equal to the other dimension. Two dimensions are considered equal if they have the
		same size and type.
		
		:param other: The other dimension.
		:type other: int or Dimension
		
		:return: True if the dimensions are equal, False otherwise.
		:rtype: bool
		"""
		if other is None:
			return False
		if isinstance(other, int):
			return self == self.from_int(other)
		return self.size == other.size and self.dtype == other.dtype

	def __int__(self) -> int:
		return self.size

	def __str__(self) -> str:
		return f"{self.name}:[{self.size}, {self.dtype.name}]"

	def __repr__(self) -> str:
		return self.__str__()

	@staticmethod
	def from_int(size: Optional[int]) -> "Dimension":
		"""
		Create a Dimension from an integer.
		
		:param size: The size of the dimension.
		:type size: int
		
		:return: A dimension with the given size and None as dtype.
		:rtype: Dimension
		"""
		return Dimension(size, DimensionProperty.NONE)

	@staticmethod
	def from_int_or_dimension(dimension: Optional[Union[int, "Dimension"]]) -> "Dimension":
		"""
		Create a Dimension from an integer or a Dimension.

		:param dimension: The dimension to convert.
		:type dimension: int or Dimension
		
		:return: A dimension with the given size and None as dtype if the input is an integer and the given dimension
		if it is a Dimension.
		:rtype: Dimension
		"""
		if isinstance(dimension, int) or dimension is None:
			return Dimension.from_int(dimension)
		return dimension


DimensionLike = Union[int, Dimension]


class Size:
	"""
	This object is used to represent the size of a space.
	
	:Attributes:
		dimensions (List[Dimension]): The dimensions of the space.
	"""

	def __init__(self, dimensions: Union[int, Dimension, Iterable[Union[int, Dimension]]]):
		"""
		Constructor for Size.
		
		:param dimensions: The dimensions of the space.
		:type dimensions: int or Dimension or Iterable[int or Dimension]
		"""
		if isinstance(dimensions, (int, Dimension)):
			dimensions = [dimensions]
		else:
			dimensions = list(dimensions)
		self.dimensions: List[Dimension] = [
			Dimension.from_int_or_dimension(dimension)
			for dimension in dimensions
		]

	def __str__(self) -> str:
		_str = "Size["
		for i, dim in enumerate(self.dimensions):
			_str += f"{i}:" + str(dim) + ", "
		return _str[:-2] + "]"

	def __repr__(self) -> str:
		return self.__str__()

	def __eq__(self, other: 'Size') -> bool:
		if other is None:
			return False
		return self.dimensions == other.dimensions

	def __getitem__(self, item):
		return self.dimensions[item]

	def __len__(self) -> int:
		return len(self.dimensions)

	def __iter__(self) -> Iterable[Dimension]:
		return iter(self.dimensions)


SizeTypes = Union[int, Dimension, Iterable[Union[int, Dimension]], Size]
DimensionsLike = Union[Dimension, Iterable[Dimension], Size]
